@@ -1,8 +1,9 @@
 #include "bamboo_tracker.hpp"
 #include <algorithm>
 #include <utility>
-#include <set>
+#include <unordered_set>
 #include <exception>
+#include <unordered_map>
 #include "commands.hpp"
 #include "io_handlers.hpp"
 #include "bank.hpp"
@@ -43,8 +44,10 @@ void BambooTracker::changeConfiguration(std::weak_ptr<Configuration> config)
 	setStreamRate(static_cast<int>(config.lock()->getSampleRate()));
 	setStreamDuration(static_cast<int>(config.lock()->getBufferLength()));
 	setMasterVolume(config.lock()->getMixerVolumeMaster());
-	setMasterVolumeFM(config.lock()->getMixerVolumeFM());
-	setMasterVolumeSSG(config.lock()->getMixerVolumeSSG());
+	if (mod_->getMixerType() == MixerType::UNSPECIFIED) {
+		setMasterVolumeFM(config.lock()->getMixerVolumeFM());
+		setMasterVolumeSSG(config.lock()->getMixerVolumeSSG());
+	}
 	playback_->setChannelRetrieving(config.lock()->getRetrieveChannelState());
 }
 
@@ -111,16 +114,16 @@ void BambooTracker::deepCloneInstrument(int num, int refNum)
 	comMan_.invoke(std::make_unique<DeepCloneInstrumentCommand>(instMan_, num, refNum));
 }
 
-void BambooTracker::loadInstrument(std::string path, int instNum)
+void BambooTracker::loadInstrument(BinaryContainer& container, std::string path, int instNum)
 {
-	auto inst = InstrumentIO::loadInstrument(path, instMan_, instNum);
+	auto inst = InstrumentIO::loadInstrument(container, path, instMan_, instNum);
 	comMan_.invoke(std::make_unique<AddInstrumentCommand>(
 					   instMan_, std::unique_ptr<AbstractInstrument>(inst)));
 }
 
-void BambooTracker::saveInstrument(std::string path, int instNum)
+void BambooTracker::saveInstrument(BinaryContainer& container, int instNum)
 {
-	InstrumentIO::saveInstrument(path, instMan_, instNum);
+	InstrumentIO::saveInstrument(container, instMan_, instNum);
 }
 
 void BambooTracker::importInstrument(const AbstractBank &bank, size_t index, int instNum)
@@ -130,9 +133,9 @@ void BambooTracker::importInstrument(const AbstractBank &bank, size_t index, int
 					   instMan_, std::unique_ptr<AbstractInstrument>(inst)));
 }
 
-void BambooTracker::exportInstruments(std::string path, std::vector<size_t> instNums)
+void BambooTracker::exportInstruments(BinaryContainer& container, std::vector<int> instNums)
 {
-	BankIO::saveBank(path, instNums, instMan_);
+	BankIO::saveBank(container, instNums, instMan_);
 }
 
 int BambooTracker::findFirstFreeInstrumentNumber() const
@@ -158,10 +161,9 @@ std::vector<int> BambooTracker::getInstrumentIndices() const
 std::vector<int> BambooTracker::getUnusedInstrumentIndices() const
 {
 	std::vector<int> unused;
-	std::set<int> regdInsts = mod_->getRegisterdInstruments();
+	std::unordered_set<int> regdInsts = mod_->getRegisterdInstruments();
 	for (auto& inst : instMan_->getInstrumentIndices()) {
-		if (regdInsts.find(inst) == regdInsts.end())
-			unused.push_back(inst);
+		if (!regdInsts.count(inst)) unused.push_back(inst);
 	}
 	return unused;
 }
@@ -176,6 +178,11 @@ std::vector<std::string> BambooTracker::getInstrumentNames() const
 	return instMan_->getInstrumentNameList();
 }
 
+std::vector<std::vector<int>> BambooTracker::checkDuplicateInstruments() const
+{
+	return instMan_->checkDuplicateInstruments();
+}
+
 //--- FM
 void BambooTracker::setEnvelopeFMParameter(int envNum, FMEnvelopeParameter param, int value)
 {
@@ -265,7 +272,7 @@ std::vector<int> BambooTracker::getOperatorSequenceFMUsers(FMEnvelopeParameter p
 	return instMan_->getOperatorSequenceFMUsers(param, opSeqNum);
 }
 
-void BambooTracker::setArpeggioFMType(int arpNum, int type)
+void BambooTracker::setArpeggioFMType(int arpNum, SequenceType type)
 {
 	instMan_->setArpeggioFMType(arpNum, type);
 }
@@ -312,7 +319,7 @@ std::vector<int> BambooTracker::getArpeggioFMUsers(int arpNum) const
 	return instMan_->getArpeggioFMUsers(arpNum);
 }
 
-void BambooTracker::setPitchFMType(int ptNum, int type)
+void BambooTracker::setPitchFMType(int ptNum, SequenceType type)
 {
 	instMan_->setPitchFMType(ptNum, type);
 }
@@ -492,7 +499,7 @@ std::vector<int> BambooTracker::getEnvelopeSSGUsers(int envNum) const
 	return instMan_->getEnvelopeSSGUsers(envNum);
 }
 
-void BambooTracker::setArpeggioSSGType(int arpNum, int type)
+void BambooTracker::setArpeggioSSGType(int arpNum, SequenceType type)
 {
 	instMan_->setArpeggioSSGType(arpNum, type);
 }
@@ -539,7 +546,7 @@ std::vector<int> BambooTracker::getArpeggioSSGUsers(int arpNum) const
 	return instMan_->getArpeggioSSGUsers(arpNum);
 }
 
-void BambooTracker::setPitchSSGType(int ptNum, int type)
+void BambooTracker::setPitchSSGType(int ptNum, SequenceType type)
 {
 	instMan_->setPitchSSGType(ptNum, type);
 }
@@ -613,7 +620,8 @@ void BambooTracker::setCurrentSongNumber(int num)
 	tickCounter_->setTempo(song.getTempo());
 	tickCounter_->setSpeed(song.getSpeed());
 	tickCounter_->setGroove(mod_->getGroove(song.getGroove()).getSequence());
-	tickCounter_->setGrooveEnebled(!song.isUsedTempo());
+	tickCounter_->setGrooveTrigger(song.isUsedTempo() ? GrooveTrigger::Invalid
+													  : GrooveTrigger::ValidByGlobal);
 
 	size_t fmch = getFMChannelCount(songStyle_.type);
 	muteStateFM_ = std::vector<bool>(fmch, false);
@@ -680,18 +688,45 @@ bool BambooTracker::isJamMode() const
 
 void BambooTracker::jamKeyOn(JamKey key)
 {
-	auto attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
-	if (attrib.source == SoundSource::Drum) {
-		opnaCtrl_->keyOnDrum(attrib.channelInSource);
+	int keyNum = octaveAndNoteToNoteNumber(octave_, JamManager::jamKeyToNote(key));
+	const TrackAttribute& attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
+	funcJamKeyOn(key, keyNum, attrib);
+}
+
+void BambooTracker::jamKeyOn(int keyNum)
+{
+	const TrackAttribute& attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
+	funcJamKeyOn(JamKey::MidiKey, keyNum, attrib);
+}
+
+void BambooTracker::jamKeyOnForced(JamKey key, SoundSource src)
+{
+	int keyNum = octaveAndNoteToNoteNumber(octave_, JamManager::jamKeyToNote(key));
+	auto it = std::find_if(songStyle_.trackAttribs.begin(), songStyle_.trackAttribs.end(),
+						   [src](TrackAttribute& attrib) { return attrib.source == src; });
+	funcJamKeyOn(key, keyNum, *it);
+}
+
+void BambooTracker::jamKeyOnForced(int keyNum, SoundSource src)
+{
+	auto it = std::find_if(songStyle_.trackAttribs.begin(), songStyle_.trackAttribs.end(),
+						   [src](TrackAttribute& attrib) { return attrib.source == src; });
+	funcJamKeyOn(JamKey::MidiKey, keyNum, *it);
+}
+
+void BambooTracker::funcJamKeyOn(JamKey key, int keyNum, const TrackAttribute& attrib)
+{
+	if (attrib.source == SoundSource::DRUM) {
+		opnaCtrl_->setKeyOnFlagDrum(attrib.channelInSource);
+		opnaCtrl_->updateRegisterStates();
 	}
 	else {
-		int keyNum = octaveAndNoteToNoteNumber(octave_, JamManager::jamKeyToNote(key));
 		std::vector<JamKeyData>&& list = jamMan_->keyOn(key, attrib.channelInSource, attrib.source, keyNum);
 		if (list.size() == 2) {	// Key off
 			JamKeyData& offData = list[1];
 			switch (offData.source) {
 			case SoundSource::FM:
-				if (songStyle_.type == SongType::FMEX && offData.channelInSource == 2) {
+				if (songStyle_.type == SongType::FM3chExpanded && offData.channelInSource == 2) {
 					opnaCtrl_->keyOffFM(2, true);
 					opnaCtrl_->keyOffFM(6, true);
 					opnaCtrl_->keyOffFM(7, true);
@@ -712,15 +747,28 @@ void BambooTracker::jamKeyOn(JamKey key)
 		std::shared_ptr<AbstractInstrument> tmpInst = instMan_->getInstrumentSharedPtr(curInstNum_);
 		JamKeyData& onData = list.front();
 
-		Note note = JamManager::jamKeyToNote(onData.key);
-		int octave = JamManager::calcOctave(octave_, onData.key);
-		int pitch = 0;
+		Note note;
+		int octave, pitch;
+		if (key == JamKey::MidiKey) {
+			auto octNote = noteNumberToOctaveAndNote(onData.keyNum);
+			note = octNote.second;
+			octave = octNote.first;
+		}
+		else {
+			note = JamManager::jamKeyToNote(onData.key);
+			octave = JamManager::calcOctave(octave_, onData.key);
+			if (octave > 7) {	// Tone range check
+				octave = 7;
+				note = Note::B;
+			}
+		}
+		pitch = 0;
 
 		switch (onData.source) {
 		case SoundSource::FM:
 			if (auto fm = std::dynamic_pointer_cast<InstrumentFM>(tmpInst))
 				opnaCtrl_->setInstrumentFM(onData.channelInSource, fm);
-			if (songStyle_.type == SongType::FMEX && onData.channelInSource == 2) {
+			if (songStyle_.type == SongType::FM3chExpanded && onData.channelInSource == 2) {
 				opnaCtrl_->keyOnFM(2, note, octave, pitch, true);
 				opnaCtrl_->keyOnFM(6, note, octave, pitch, true);
 				opnaCtrl_->keyOnFM(7, note, octave, pitch, true);
@@ -741,113 +789,47 @@ void BambooTracker::jamKeyOn(JamKey key)
 	}
 }
 
-void BambooTracker::jamKeyOn(int keyNum)
-{
-	auto attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
-	if (attrib.source == SoundSource::Drum) {
-		opnaCtrl_->keyOnDrum(attrib.channelInSource);
-	}
-	else {
-		std::vector<JamKeyData>&& list = jamMan_->keyOn(JamKey::MIDI_KEY, attrib.channelInSource, attrib.source, keyNum);
-		if (list.size() == 2) {	// Key off
-			JamKeyData& offData = list[1];
-			switch (offData.source) {
-			case SoundSource::FM:
-				if (songStyle_.type == SongType::FMEX && offData.channelInSource == 2) {
-					opnaCtrl_->keyOffFM(2, true);
-					opnaCtrl_->keyOffFM(6, true);
-					opnaCtrl_->keyOffFM(7, true);
-					opnaCtrl_->keyOffFM(8, true);
-				}
-				else {
-					opnaCtrl_->keyOffFM(offData.channelInSource, true);
-				}
-				break;
-			case SoundSource::SSG:
-				opnaCtrl_->keyOffSSG(offData.channelInSource, true);
-				break;
-			default:
-				break;
-			}
-		}
-
-		std::shared_ptr<AbstractInstrument> tmpInst = instMan_->getInstrumentSharedPtr(curInstNum_);
-		JamKeyData& onData = list.front();
-
-		auto octNote = noteNumberToOctaveAndNote(keyNum);
-		int pitch = 0;
-
-		switch (onData.source) {
-		case SoundSource::FM:
-			if (auto fm = std::dynamic_pointer_cast<InstrumentFM>(tmpInst))
-				opnaCtrl_->setInstrumentFM(onData.channelInSource, fm);
-			if (songStyle_.type == SongType::FMEX && onData.channelInSource == 2) {
-				opnaCtrl_->keyOnFM(2, octNote.second, octNote.first, pitch, true);
-				opnaCtrl_->keyOnFM(6, octNote.second, octNote.first, pitch, true);
-				opnaCtrl_->keyOnFM(7, octNote.second, octNote.first, pitch, true);
-				opnaCtrl_->keyOnFM(8, octNote.second, octNote.first, pitch, true);
-			}
-			else {
-				opnaCtrl_->keyOnFM(onData.channelInSource, octNote.second, octNote.first, pitch, true);
-			}
-			break;
-		case SoundSource::SSG:
-			if (auto ssg = std::dynamic_pointer_cast<InstrumentSSG>(tmpInst))
-				opnaCtrl_->setInstrumentSSG(onData.channelInSource, ssg);
-			opnaCtrl_->keyOnSSG(onData.channelInSource, octNote.second, octNote.first, pitch, true);
-			break;
-		default:
-			break;
-		}
-	}
-}
-
 void BambooTracker::jamKeyOff(JamKey key)
 {
-	auto attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
-	if (attrib.source == SoundSource::Drum) {
-		opnaCtrl_->keyOffDrum(attrib.channelInSource);
+	int keyNum = octaveAndNoteToNoteNumber(octave_, JamManager::jamKeyToNote(key));
+	const TrackAttribute& attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
+	funcJamKeyOff(key, keyNum, attrib);
+}
+
+void BambooTracker::jamKeyOff(int keyNum)
+{
+	const TrackAttribute& attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
+	funcJamKeyOff(JamKey::MidiKey, keyNum, attrib);
+}
+
+void BambooTracker::jamKeyOffForced(JamKey key, SoundSource src)
+{
+	int keyNum = octaveAndNoteToNoteNumber(octave_, JamManager::jamKeyToNote(key));
+	auto it = std::find_if(songStyle_.trackAttribs.begin(), songStyle_.trackAttribs.end(),
+						   [src](TrackAttribute& attrib) { return attrib.source == src; });
+	funcJamKeyOff(key, keyNum, *it);
+}
+
+void BambooTracker::jamKeyOffForced(int keyNum, SoundSource src)
+{
+	auto it = std::find_if(songStyle_.trackAttribs.begin(), songStyle_.trackAttribs.end(),
+						   [src](TrackAttribute& attrib) { return attrib.source == src; });
+	funcJamKeyOff(JamKey::MidiKey, keyNum, *it);
+}
+
+void BambooTracker::funcJamKeyOff(JamKey key, int keyNum, const TrackAttribute& attrib)
+{
+	if (attrib.source == SoundSource::DRUM) {
+		opnaCtrl_->setKeyOffFlagDrum(attrib.channelInSource);
+		opnaCtrl_->updateRegisterStates();
 	}
 	else {
-		int keyNum = octaveAndNoteToNoteNumber(octave_, JamManager::jamKeyToNote(key));
 		JamKeyData&& data = jamMan_->keyOff(key, keyNum);
 
 		if (data.channelInSource > -1) {	// Key still sound
 			switch (data.source) {
 			case SoundSource::FM:
-				if (songStyle_.type == SongType::FMEX && data.channelInSource == 2) {
-					opnaCtrl_->keyOffFM(2, true);
-					opnaCtrl_->keyOffFM(6, true);
-					opnaCtrl_->keyOffFM(7, true);
-					opnaCtrl_->keyOffFM(8, true);
-				}
-				else {
-					opnaCtrl_->keyOffFM(data.channelInSource, true);
-				}
-				break;
-			case SoundSource::SSG:
-				opnaCtrl_->keyOffSSG(data.channelInSource, true);
-				break;
-			default:
-				break;
-			}
-		}
-	}
-}
-
-void BambooTracker::jamKeyOff(int keyNum)
-{
-	auto attrib = songStyle_.trackAttribs[static_cast<size_t>(curTrackNum_)];
-	if (attrib.source == SoundSource::Drum) {
-		opnaCtrl_->keyOffDrum(attrib.channelInSource);
-	}
-	else {
-		JamKeyData&& data = jamMan_->keyOff(JamKey::MIDI_KEY, keyNum);
-
-		if (data.channelInSource > -1) {	// Key still sound
-			switch (data.source) {
-			case SoundSource::FM:
-				if (songStyle_.type == SongType::FMEX && data.channelInSource == 2) {
+				if (songStyle_.type == SongType::FM3chExpanded && data.channelInSource == 2) {
 					opnaCtrl_->keyOffFM(2, true);
 					opnaCtrl_->keyOffFM(6, true);
 					opnaCtrl_->keyOffFM(7, true);
@@ -925,6 +907,11 @@ bool BambooTracker::isPlaySong() const
 	return playback_->isPlaySong();
 }
 
+PlaybackState BambooTracker::getPlaybackState() const
+{
+	return playback_->getPlaybackState();
+}
+
 void BambooTracker::setTrackMuteState(int trackNum, bool isMute)
 {
 	auto& ta = songStyle_.trackAttribs[static_cast<size_t>(trackNum)];
@@ -938,7 +925,7 @@ void BambooTracker::setTrackMuteState(int trackNum, bool isMute)
 		muteStateSSG_.at(static_cast<size_t>(ch)) = isMute;
 		if (isPlaySong()) opnaCtrl_->setMuteSSGState(ch, isMute);
 		break;
-	case SoundSource::Drum:
+	case SoundSource::DRUM:
 		muteStateDrum_.at(static_cast<size_t>(ch)) = isMute;
 		if (isPlaySong()) opnaCtrl_->setMuteDrumState(ch, isMute);
 		break;
@@ -952,7 +939,7 @@ bool BambooTracker::isMute(int trackNum)
 	switch (ta.source) {
 	case SoundSource::FM:	return muteStateFM_.at(ch);
 	case SoundSource::SSG:	return muteStateSSG_.at(ch);
-	case SoundSource::Drum:	return muteStateDrum_.at(ch);
+	case SoundSource::DRUM:	return muteStateDrum_.at(ch);
 	default:	return false;	// Dummy
 	}
 }
@@ -960,6 +947,13 @@ bool BambooTracker::isMute(int trackNum)
 void BambooTracker::setFollowPlay(bool isFollowed)
 {
 	isFollowPlay_ = isFollowed;
+	if (isFollowed) {
+		int odr = playback_->getPlayingOrderNumber();
+		if (odr >= 0) {
+			curOrderNum_ = odr;
+			curStepNum_ = playback_->getPlayingStepNumber();
+		}
+	}
 }
 
 bool BambooTracker::isFollowPlay() const
@@ -978,8 +972,10 @@ int BambooTracker::getPlayingStepNumber() const
 }
 
 /********** Export **********/
-bool BambooTracker::exportToWav(std::string file, int loopCnt, std::function<bool()> f)
+bool BambooTracker::exportToWav(BinaryContainer& container, int rate, int loopCnt, std::function<bool()> bar)
 {
+	int tmpRate = opnaCtrl_->getRate();
+	opnaCtrl_->setRate(rate);
 	size_t sampCnt = static_cast<size_t>(opnaCtrl_->getRate() * opnaCtrl_->getDuration() / 1000);
 	size_t intrCnt = static_cast<size_t>(opnaCtrl_->getRate()) / mod_->getTickFrequency();
 	size_t intrCntRest = 0;
@@ -1001,7 +997,7 @@ bool BambooTracker::exportToWav(std::string file, int loopCnt, std::function<boo
 				intrCntRest = intrCnt;    // Set counts to next interruption
 
 				if (!streamCountUp()) {
-					if (f()) {	// Update lambda function
+					if (bar()) {	// Update lambda function
 						stopPlaySong();
 						isFollowPlay_ = tmpFollow;
 						return false;
@@ -1030,10 +1026,10 @@ bool BambooTracker::exportToWav(std::string file, int loopCnt, std::function<boo
 	opnaCtrl_->setExportContainer();
 	stopPlaySong();
 	isFollowPlay_ = tmpFollow;
+	opnaCtrl_->setRate(tmpRate);
 
 	try {
-		ExportHandler::writeWave(file, exCntr->getStream(), static_cast<uint32_t>(opnaCtrl_->getRate()));
-		f();
+		ExportHandler::writeWave(container, exCntr->getStream(), static_cast<uint32_t>(rate));
 		return true;
 	}
 	catch (...) {
@@ -1041,7 +1037,8 @@ bool BambooTracker::exportToWav(std::string file, int loopCnt, std::function<boo
 	}
 }
 
-bool BambooTracker::exportToVgm(std::string file, int target, bool gd3TagEnabled, GD3Tag tag, std::function<bool()> f)
+bool BambooTracker::exportToVgm(BinaryContainer& container, int target, bool gd3TagEnabled,
+								GD3Tag tag, std::function<bool()> bar)
 {
 	int tmpRate = opnaCtrl_->getRate();
 	opnaCtrl_->setRate(44100);
@@ -1067,7 +1064,7 @@ bool BambooTracker::exportToVgm(std::string file, int target, bool gd3TagEnabled
 
 	while (true) {
 		if (!streamCountUp()) {
-			if (f()) {	// Update lambda function
+			if (bar()) {	// Update lambda function
 				stopPlaySong();
 				isFollowPlay_ = tmpFollow;
 				return false;
@@ -1095,17 +1092,17 @@ bool BambooTracker::exportToVgm(std::string file, int target, bool gd3TagEnabled
 	opnaCtrl_->setRate(tmpRate);
 
 	try {
-		ExportHandler::writeVgm(file, target, exCntr->getData(), CHIP_CLOCK, mod_->getTickFrequency(),
+		ExportHandler::writeVgm(container, target, exCntr->getData(), CHIP_CLOCK, mod_->getTickFrequency(),
 								loopFlag, loopPoint, exCntr->getSampleLength() - loopPointSamples,
 								exCntr->getSampleLength(), gd3TagEnabled, tag);
-		f();
 		return true;
 	} catch (...) {
 		throw;
 	}
 }
 
-bool BambooTracker::exportToS98(std::string file, int target, bool tagEnabled, S98Tag tag, int rate, std::function<bool()> f)
+bool BambooTracker::exportToS98(BinaryContainer& container, int target, bool tagEnabled,
+								S98Tag tag, int rate, std::function<bool()> bar)
 {
 	int tmpRate = opnaCtrl_->getRate();
 	opnaCtrl_->setRate(rate);
@@ -1130,7 +1127,7 @@ bool BambooTracker::exportToS98(std::string file, int target, bool tagEnabled, S
 	while (true) {
 		exCntr->getData();	// Set wait counts
 		if (!streamCountUp()) {
-			if (f()) {	// Update lambda function
+			if (bar()) {	// Update lambda function
 				stopPlaySong();
 				isFollowPlay_ = tmpFollow;
 				return false;
@@ -1157,9 +1154,8 @@ bool BambooTracker::exportToS98(std::string file, int target, bool tagEnabled, S
 	opnaCtrl_->setRate(tmpRate);
 
 	try {
-		ExportHandler::writeS98(file, target, exCntr->getData(), CHIP_CLOCK, static_cast<uint32_t>(rate),
+		ExportHandler::writeS98(container, target, exCntr->getData(), CHIP_CLOCK, static_cast<uint32_t>(rate),
 								loopFlag, loopPoint, tagEnabled, tag);
-		f();
 		return true;
 	} catch (...) {
 		throw;
@@ -1201,12 +1197,6 @@ void BambooTracker::checkNextPositionOfLastStep(int& endOrder, int& endStep) con
 	}
 }
 
-/********** Backup **********/
-void BambooTracker::backupModule(std::string file)
-{
-	ModuleIO::backupModule(file);
-}
-
 /********** Stream type **********/
 void BambooTracker::useSCCI(SoundInterfaceManager* manager)
 {
@@ -1309,13 +1299,13 @@ void BambooTracker::makeNewModule()
 	clearCommandHistory();
 }
 
-void BambooTracker::loadModule(std::string path)
+void BambooTracker::loadModule(BinaryContainer& container)
 {
 	makeNewModule();
 
 	std::exception_ptr ep;
 	try {
-		ModuleIO::loadModule(path, mod_, instMan_);
+		ModuleIO::loadModule(container, mod_, instMan_);
 	}
 	catch (...) {
 		ep = std::current_exception();
@@ -1328,9 +1318,9 @@ void BambooTracker::loadModule(std::string path)
 	if (ep) std::rethrow_exception(ep);
 }
 
-void BambooTracker::saveModule(std::string path)
+void BambooTracker::saveModule(BinaryContainer& container)
 {
-	ModuleIO::saveModule(path, mod_, instMan_);
+	ModuleIO::saveModule(container, mod_, instMan_);
 }
 
 void BambooTracker::setModulePath(std::string path)
@@ -1414,6 +1404,36 @@ size_t BambooTracker::getModuleStepHighlight2Distance() const
 	return mod_->getStepHighlight2Distance();
 }
 
+void BambooTracker::setModuleMixerType(MixerType type)
+{
+	mod_->setMixerType(type);
+}
+
+MixerType BambooTracker::getModuleMixerType() const
+{
+	return mod_->getMixerType();
+}
+
+void BambooTracker::setModuleCustomMixerFMLevel(double level)
+{
+	mod_->setCustomMixerFMLevel(level);
+}
+
+double BambooTracker::getModuleCustomMixerFMLevel() const
+{
+	return mod_->getCustomMixerFMLevel();
+}
+
+void BambooTracker::setModuleCustomMixerSSGLevel(double level)
+{
+	mod_->setCustomMixerSSGLevel(level);
+}
+
+double BambooTracker::getModuleCustomMixerSSGLevel() const
+{
+	return mod_->getCustomMixerSSGLevel();
+}
+
 size_t BambooTracker::getGrooveCount() const
 {
 	return mod_->getGrooveCount();
@@ -1439,6 +1459,18 @@ void BambooTracker::clearUnusedPatterns()
 	mod_->clearUnusedPatterns();
 }
 
+void BambooTracker::replaceDuplicateInstrumentsInPatterns(std::vector<std::vector<int>> list)
+{
+	std::unordered_map<int, int> map;
+	for (auto& group : list) {
+		for (size_t i = 1; i < group.size(); ++i) {
+			map.emplace(group[i], group.front());
+		}
+	}
+
+	mod_->replaceDuplicateInstrumentsInPatterns(map);
+}
+
 /*----- Song -----*/
 void BambooTracker::setSongTitle(int songNum, std::string title)
 {
@@ -1475,7 +1507,8 @@ int BambooTracker::getSongGroove(int songNum) const
 void BambooTracker::toggleTempoOrGrooveInSong(int songNum, bool isTempo)
 {
 	mod_->getSong(songNum).toggleTempoOrGroove(isTempo);
-	tickCounter_->setGrooveEnebled(!isTempo);
+	tickCounter_->setGrooveTrigger(isTempo ? GrooveTrigger::Invalid
+										   : GrooveTrigger::ValidByGlobal);
 }
 
 bool BambooTracker::isUsedTempoInSong(int songNum) const
@@ -1556,9 +1589,9 @@ std::vector<OrderData> BambooTracker::getOrderData(int songNum, int orderNum) co
 	return mod_->getSong(songNum).getOrderData(orderNum);
 }
 
-void BambooTracker::setOrderPattern(int songNum, int trackNum, int orderNum, int patternNum)
+void BambooTracker::setOrderPatternDigit(int songNum, int trackNum, int orderNum, int patternNum, bool secondEntry)
 {
-	comMan_.invoke(std::make_unique<SetPatternToOrderCommand>(mod_, songNum, trackNum, orderNum, patternNum));
+	comMan_.invoke(std::make_unique<SetPatternToOrderCommand>(mod_, songNum, trackNum, orderNum, patternNum, secondEntry));
 }
 
 void BambooTracker::insertOrderBelow(int songNum, int orderNum)
@@ -1629,19 +1662,17 @@ int BambooTracker::getStepNoteNumber(int songNum, int trackNum, int orderNum, in
 			.getStep(stepNum).getNoteNumber();
 }
 
-void BambooTracker::setStepNote(int songNum, int trackNum, int orderNum, int stepNum, int octave, Note note)
+void BambooTracker::setStepNote(int songNum, int trackNum, int orderNum, int stepNum, int octave, Note note, bool autosetInst)
 {
 	int nn = octaveAndNoteToNoteNumber(octave, note);
 
-	int in;
-	if (curInstNum_ != -1
+	int in = -1;
+	if (autosetInst && curInstNum_ != -1
 			&& (songStyle_.trackAttribs.at(static_cast<size_t>(trackNum)).source
 				== instMan_->getInstrumentSharedPtr(curInstNum_)->getSoundSource()))
 		in = curInstNum_;
-	else
-		in = -1;
 
-	comMan_.invoke(std::make_unique<SetKeyOnToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, nn, in));
+	comMan_.invoke(std::make_unique<SetKeyOnToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, nn, autosetInst, in));
 }
 
 void BambooTracker::setStepKeyOff(int songNum, int trackNum, int orderNum, int stepNum)
@@ -1665,9 +1696,9 @@ int BambooTracker::getStepInstrument(int songNum, int trackNum, int orderNum, in
 			.getStep(stepNum).getInstrumentNumber();
 }
 
-void BambooTracker::setStepInstrument(int songNum, int trackNum, int orderNum, int stepNum, int instNum)
+void BambooTracker::setStepInstrumentDigit(int songNum, int trackNum, int orderNum, int stepNum, int instNum, bool secondEntry)
 {
-	comMan_.invoke(std::make_unique<SetInstrumentToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, instNum));
+	comMan_.invoke(std::make_unique<SetInstrumentToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, instNum, secondEntry));
 }
 
 void BambooTracker::eraseStepInstrument(int songNum, int trackNum, int orderNum, int stepNum)
@@ -1681,9 +1712,9 @@ int BambooTracker::getStepVolume(int songNum, int trackNum, int orderNum, int st
 			.getStep(stepNum).getVolume();
 }
 
-void BambooTracker::setStepVolume(int songNum, int trackNum, int orderNum, int stepNum, int volume, bool isFMReversed)
+void BambooTracker::setStepVolumeDigit(int songNum, int trackNum, int orderNum, int stepNum, int volume, bool isFMReversed, bool secondEntry)
 {	
-	comMan_.invoke(std::make_unique<SetVolumeToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, volume, isFMReversed));
+	comMan_.invoke(std::make_unique<SetVolumeToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, volume, isFMReversed, secondEntry));
 }
 
 void BambooTracker::eraseStepVolume(int songNum, int trackNum, int orderNum, int stepNum)
@@ -1697,9 +1728,9 @@ std::string BambooTracker::getStepEffectID(int songNum, int trackNum, int orderN
 			.getStep(stepNum).getEffectID(n);
 }
 
-void BambooTracker::setStepEffectID(int songNum, int trackNum, int orderNum, int stepNum, int n, std::string id, bool fillValue00)
+void BambooTracker::setStepEffectIDCharacter(int songNum, int trackNum, int orderNum, int stepNum, int n, std::string id, bool fillValue00, bool secondEntry)
 {
-	comMan_.invoke(std::make_unique<SetEffectIDToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, n, id, fillValue00));
+	comMan_.invoke(std::make_unique<SetEffectIDToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, n, id, fillValue00, secondEntry));
 }
 
 int BambooTracker::getStepEffectValue(int songNum, int trackNum, int orderNum, int stepNum, int n) const
@@ -1708,9 +1739,9 @@ int BambooTracker::getStepEffectValue(int songNum, int trackNum, int orderNum, i
 			.getStep(stepNum).getEffectValue(n);
 }
 
-void BambooTracker::setStepEffectValue(int songNum, int trackNum, int orderNum, int stepNum, int n, int value, bool isFMReversed)
+void BambooTracker::setStepEffectValueDigit(int songNum, int trackNum, int orderNum, int stepNum, int n, int value, EffectDisplayControl ctrl, bool secondEntry)
 {
-	comMan_.invoke(std::make_unique<SetEffectValueToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, n, value, isFMReversed));
+	comMan_.invoke(std::make_unique<SetEffectValueToStepCommand>(mod_, songNum, trackNum, orderNum, stepNum, n, value, ctrl, secondEntry));
 }
 
 void BambooTracker::eraseStepEffect(int songNum, int trackNum, int orderNum, int stepNum, int n)
