@@ -19,6 +19,7 @@
 #include "chips/scci/scci.h"
 #include "effect.hpp"
 #include "playback.hpp"
+#include "binary_container.hpp"
 #include "misc.hpp"
 
 class AbstractBank;
@@ -49,10 +50,10 @@ public:
 	std::unique_ptr<AbstractInstrument> getInstrument(int num);
 	void cloneInstrument(int num, int refNum);
 	void deepCloneInstrument(int num, int refNum);
-	void loadInstrument(std::string path, int instNum);
-	void saveInstrument(std::string path, int instNum);
+	void loadInstrument(BinaryContainer& container, std::string path, int instNum);
+	void saveInstrument(BinaryContainer& container, int instNum);
 	void importInstrument(const AbstractBank &bank, size_t index, int instNum);
-	void exportInstruments(std::string path, std::vector<size_t> instNums);
+	void exportInstruments(BinaryContainer& container, std::vector<int> instNums);
 	int findFirstFreeInstrumentNumber() const;
 	void setInstrumentName(int num, std::string name);
 	void clearAllInstrument();
@@ -60,6 +61,7 @@ public:
 	std::vector<int> getUnusedInstrumentIndices() const;
 	void clearUnusedInstrumentProperties();
 	std::vector<std::string> getInstrumentNames() const;
+	std::vector<std::vector<int>> checkDuplicateInstruments() const;
 
 	//--- FM
 	void setEnvelopeFMParameter(int envNum, FMEnvelopeParameter param, int value);
@@ -81,7 +83,7 @@ public:
 	void setInstrumentFMOperatorSequenceEnabled(int instNum, FMEnvelopeParameter param, bool enabled);
 	std::vector<int> getOperatorSequenceFMUsers(FMEnvelopeParameter param, int opSeqNum) const;
 
-	void setArpeggioFMType(int arpNum, int type);
+	void setArpeggioFMType(int arpNum, SequenceType type);
 	void addArpeggioFMSequenceCommand(int arpNum, int type, int data);
 	void removeArpeggioFMSequenceCommand(int arpNum);
 	void setArpeggioFMSequenceCommand(int arpNum, int cnt, int type, int data);
@@ -91,7 +93,7 @@ public:
 	void setInstrumentFMArpeggioEnabled(int instNum, FMOperatorType op, bool enabled);
 	std::vector<int> getArpeggioFMUsers(int arpNum) const;
 
-	void setPitchFMType(int ptNum, int type);
+	void setPitchFMType(int ptNum, SequenceType type);
 	void addPitchFMSequenceCommand(int ptNum, int type, int data);
 	void removePitchFMSequenceCommand(int ptNum);
 	void setPitchFMSequenceCommand(int ptNum, int cnt, int type, int data);
@@ -131,7 +133,7 @@ public:
 	void setInstrumentSSGEnvelopeEnabled(int instNum, bool enabled);
 	std::vector<int> getEnvelopeSSGUsers(int envNum) const;
 
-	void setArpeggioSSGType(int arpNum, int type);
+	void setArpeggioSSGType(int arpNum, SequenceType type);
 	void addArpeggioSSGSequenceCommand(int arpNum, int type, int data);
 	void removeArpeggioSSGSequenceCommand(int arpNum);
 	void setArpeggioSSGSequenceCommand(int arpNum, int cnt, int type, int data);
@@ -141,7 +143,7 @@ public:
 	void setInstrumentSSGArpeggioEnabled(int instNum, bool enabled);
 	std::vector<int> getArpeggioSSGUsers(int arpNum) const;
 
-	void setPitchSSGType(int ptNum, int type);
+	void setPitchSSGType(int ptNum, SequenceType type);
 	void addPitchSSGSequenceCommand(int ptNum, int type, int data);
 	void removePitchSSGSequenceCommand(int ptNum);
 	void setPitchSSGSequenceCommand(int ptNum, int cnt, int type, int data);
@@ -175,6 +177,10 @@ public:
 	void jamKeyOn(int keyNum);
 	void jamKeyOff(JamKey key);
 	void jamKeyOff(int keyNum);
+	void jamKeyOnForced(JamKey key, SoundSource src);
+	void jamKeyOnForced(int keyNum, SoundSource src);
+	void jamKeyOffForced(JamKey key, SoundSource src);
+	void jamKeyOffForced(int keyNum, SoundSource src);
 
 	// Play song
 	void startPlaySong();
@@ -183,6 +189,7 @@ public:
 	void startPlayFromCurrentStep();
 	void stopPlaySong();
 	bool isPlaySong() const;
+	PlaybackState getPlaybackState() const;
 	void setTrackMuteState(int trackNum, bool isMute);
 	bool isMute(int trackNum);
 	void setFollowPlay(bool isFollowed);
@@ -191,12 +198,11 @@ public:
 	int getPlayingStepNumber() const;
 
 	// Export
-	bool exportToWav(std::string file, int loopCnt, std::function<bool()> f);
-	bool exportToVgm(std::string file, int target, bool gd3TagEnabled, GD3Tag tag, std::function<bool()> f);
-	bool exportToS98(std::string file, int target, bool tagEnabled, S98Tag tag, int rate, std::function<bool()> f);
-
-	// Backup
-	void backupModule(std::string file);
+	bool exportToWav(BinaryContainer& container, int rate, int loopCnt, std::function<bool()> bar);
+	bool exportToVgm(BinaryContainer& container, int target, bool gd3TagEnabled,
+					 GD3Tag tag, std::function<bool()> bar);
+	bool exportToS98(BinaryContainer& container, int target, bool tagEnabled, S98Tag tag,
+					 int rate, std::function<bool()> bar);
 
 	// Stream type
 	void useSCCI(SoundInterfaceManager* manager);
@@ -224,8 +230,8 @@ public:
 	// Module details
 	/*----- Module -----*/
 	void makeNewModule();
-	void loadModule(std::string path);
-	void saveModule(std::string path);
+	void loadModule(BinaryContainer& container);
+	void saveModule(BinaryContainer& container);
 	void setModulePath(std::string path);
 	std::string getModulePath() const;
 	void setModuleTitle(std::string title);
@@ -242,11 +248,18 @@ public:
 	size_t getModuleStepHighlight1Distance() const;
 	void setModuleStepHighlight2Distance(size_t dist);
 	size_t getModuleStepHighlight2Distance() const;
+	void setModuleMixerType(MixerType type);
+	MixerType getModuleMixerType() const;
+	void setModuleCustomMixerFMLevel(double level);
+	double getModuleCustomMixerFMLevel() const;
+	void setModuleCustomMixerSSGLevel(double level);
+	double getModuleCustomMixerSSGLevel() const;
 	size_t getGrooveCount() const;
 	void setGroove(int num, std::vector<int> seq);
 	void setGrooves(std::vector<std::vector<int>> seqs);
 	std::vector<int> getGroove(int num) const;
 	void clearUnusedPatterns();
+	void replaceDuplicateInstrumentsInPatterns(std::vector<std::vector<int>> list);
 	/*----- Song -----*/
 	void setSongTitle(int songNum, std::string title);
 	std::string getSongTitle(int songNum) const;
@@ -268,7 +281,7 @@ public:
 	size_t getEffectDisplayWidth(int songNum, int trackNum) const;
 	/*----- Order -----*/
 	std::vector<OrderData> getOrderData(int songNum, int orderNum) const;
-	void setOrderPattern(int songNum, int trackNum, int orderNum, int patternNum);
+	void setOrderPatternDigit(int songNum, int trackNum, int orderNum, int patternNum, bool secondEntry);
 	void insertOrderBelow(int songNum, int orderNum);
 	void deleteOrder(int songNum, int orderNum);
 	void pasteOrderCells(int songNum, int beginTrack, int beginOrder,
@@ -281,20 +294,20 @@ public:
 	bool canAddNewOrder(int songNum) const;
 	/*----- Pattern -----*/
 	int getStepNoteNumber(int songNum, int trackNum, int orderNum, int stepNum) const;
-	void setStepNote(int songNum, int trackNum, int orderNum, int stepNum, int octave, Note note);
+	void setStepNote(int songNum, int trackNum, int orderNum, int stepNum, int octave, Note note, bool autosetInst);
 	void setStepKeyOff(int songNum, int trackNum, int orderNum, int stepNum);
 	void setEchoBufferAccess(int songNum, int trackNum, int orderNum, int stepNum, int bufNum);
 	void eraseStepNote(int songNum, int trackNum, int orderNum, int stepNum);
 	int getStepInstrument(int songNum, int trackNum, int orderNum, int stepNum) const;
-	void setStepInstrument(int songNum, int trackNum, int orderNum, int stepNum, int instNum);
+	void setStepInstrumentDigit(int songNum, int trackNum, int orderNum, int stepNum, int instNum, bool secondEntry);
 	void eraseStepInstrument(int songNum, int trackNum, int orderNum, int stepNum);
 	int getStepVolume(int songNum, int trackNum, int orderNum, int stepNum) const;
-	void setStepVolume(int songNum, int trackNum, int orderNum, int stepNum, int volume, bool isFMReversed);
+	void setStepVolumeDigit(int songNum, int trackNum, int orderNum, int stepNum, int volume, bool isFMReversed, bool secondEntry);
 	void eraseStepVolume(int songNum, int trackNum, int orderNum, int stepNum);
 	std::string getStepEffectID(int songNum, int trackNum, int orderNum, int stepNum, int n) const;
-	void setStepEffectID(int songNum, int trackNum, int orderNum, int stepNum, int n, std::string id, bool fillValue00);
+	void setStepEffectIDCharacter(int songNum, int trackNum, int orderNum, int stepNum, int n, std::string id, bool fillValue00, bool secondEntry);
 	int getStepEffectValue(int songNum, int trackNum, int orderNum, int stepNum, int n) const;
-	void setStepEffectValue(int songNum, int trackNum, int orderNum, int stepNum, int n, int value, bool isFMReversed);
+	void setStepEffectValueDigit(int songNum, int trackNum, int orderNum, int stepNum, int n, int value, EffectDisplayControl ctrl, bool secondEntry);
 	void eraseStepEffect(int songNum, int trackNum, int orderNum, int stepNum, int n);
 	void eraseStepEffectValue(int songNum, int trackNum, int orderNum, int stepNum, int n);
 	void deletePreviousStep(int songNum, int trackNum, int orderNum, int stepNum);
@@ -362,6 +375,10 @@ private:
 
 	static const uint32_t CHIP_CLOCK;
 
+	// Jam mode
+	void funcJamKeyOn(JamKey key, int keyNum, const TrackAttribute& attrib);
+	void funcJamKeyOff(JamKey key, int keyNum, const TrackAttribute& attrib);
+
 	// Play song
 	void startPlay();
 
