@@ -9,8 +9,9 @@ Configuration::Configuration()
 	mainMax_ = false;
 	mainX_ = -1;	// Dummy
 	mainY_ = -1;	// Dummy
+	mainVSplit_ = -1;	// Dummy
 	instFMW_ = 570;
-	instFMH_ = 680;
+	instFMH_ = 750;
 	instSSGW_ = 500;
 	instSSGH_ = 390;
 	followMode_ = true;
@@ -33,6 +34,8 @@ Configuration::Configuration()
 	showFMDetuneSigned_ = false;
 	showWaveVisual_ = true;
 	fill00ToEffectValue_ = true;
+	autosetInstrument_ = true;
+	moveCursorHScroll_ = true;
 
 	// Edit settings
 	pageJumpLength_ = 4;
@@ -47,9 +50,10 @@ Configuration::Configuration()
 	noteEntryLayout_ = QWERTY;
 
 	// Sound //
+	sndAPI_ = u8"";
 	sndDevice_ = u8"";
 	useSCCI_ = false;
-	emulator_ = 0;
+	emulator_ = 1;
 	sampleRate_ = 44100;
 	bufferLength_ = 40;
 
@@ -478,112 +482,122 @@ Configuration::Configuration()
 	};
 
 	// Layouts
-	const std::map<std::string, JamKey> mappingQWERTY = {
-		{u8"Z",         JamKey::LOW_C},
-		{u8"S",         JamKey::LOW_CS},
-		{u8"X",         JamKey::LOW_D},
-		{u8"D",         JamKey::LOW_DS},
-		{u8"C",         JamKey::LOW_E},
-		{u8"V",         JamKey::LOW_F},
-		{u8"G",         JamKey::LOW_FS},
-		{u8"B",         JamKey::LOW_G},
-		{u8"H",         JamKey::LOW_GS},
-		{u8"N",         JamKey::LOW_A},
-		{u8"J",         JamKey::LOW_AS},
-		{u8"M",         JamKey::LOW_B},
-		{u8",",         JamKey::LOW_C_H},
-		{u8"L",         JamKey::LOW_CS_H},
-		{u8".",         JamKey::LOW_D_H},
+	const std::unordered_map<std::string, JamKey> mappingQWERTY = {
+		{u8"Z",         JamKey::LowC},
+		{u8"S",         JamKey::LowCS},
+		{u8"X",         JamKey::LowD},
+		{u8"D",         JamKey::LowDS},
+		{u8"C",         JamKey::LowE},
+		{u8"V",         JamKey::LowF},
+		{u8"G",         JamKey::LowFS},
+		{u8"B",         JamKey::LowG},
+		{u8"H",         JamKey::LowGS},
+		{u8"N",         JamKey::LowA},
+		{u8"J",         JamKey::LowAS},
+		{u8"M",         JamKey::LowB},
+		{u8",",         JamKey::LowC2},
+		{u8"L",         JamKey::LowCS2},
+		{u8".",         JamKey::LowD2},
 
-		{u8"Q",         JamKey::HIGH_C},
-		{u8"2",         JamKey::HIGH_CS},
-		{u8"W",         JamKey::HIGH_D},
-		{u8"3",         JamKey::HIGH_DS},
-		{u8"E",         JamKey::HIGH_E},
-		{u8"R",         JamKey::HIGH_F},
-		{u8"5",         JamKey::HIGH_FS},
-		{u8"T",         JamKey::HIGH_G},
-		{u8"6",         JamKey::HIGH_GS},
-		{u8"Y",         JamKey::HIGH_A},
-		{u8"7",         JamKey::HIGH_AS},
-		{u8"U",         JamKey::HIGH_B},
-		{u8"I",         JamKey::HIGH_C_H},
-		{u8"9",         JamKey::HIGH_CS_H},
-		{u8"O",         JamKey::HIGH_D_H},
+		{u8"Q",         JamKey::HighC},
+		{u8"2",         JamKey::HighCS},
+		{u8"W",         JamKey::HighD},
+		{u8"3",         JamKey::HighDS},
+		{u8"E",         JamKey::HighE},
+		{u8"R",         JamKey::HighF},
+		{u8"5",         JamKey::HighFS},
+		{u8"T",         JamKey::HighG},
+		{u8"6",         JamKey::HighGS},
+		{u8"Y",         JamKey::HighA},
+		{u8"7",         JamKey::HighAS},
+		{u8"U",         JamKey::HighB},
+		{u8"I",         JamKey::HighC2},
+		{u8"9",         JamKey::HighCS2},
+		{u8"O",         JamKey::HighD2},
 	};
-	const std::map<std::string, JamKey> mappingQWERTZ = {
-		{u8"Y",         JamKey::LOW_C},
-		{u8"S",         JamKey::LOW_CS},
-		{u8"X",         JamKey::LOW_D},
-		{u8"D",         JamKey::LOW_DS},
-		{u8"C",         JamKey::LOW_E},
-		{u8"V",         JamKey::LOW_F},
-		{u8"G",         JamKey::LOW_FS},
-		{u8"B",         JamKey::LOW_G},
-		{u8"H",         JamKey::LOW_GS},
-		{u8"N",         JamKey::LOW_A},
-		{u8"J",         JamKey::LOW_AS},
-		{u8"M",         JamKey::LOW_B},
-		{u8",",         JamKey::LOW_C_H},
-		{u8"L",         JamKey::LOW_CS_H},
-		{u8".",         JamKey::LOW_D_H},
+	const std::unordered_map<std::string, JamKey> mappingQWERTZ = {
+		{u8"Y",         JamKey::LowC},
+		{u8"S",         JamKey::LowCS},
+		{u8"X",         JamKey::LowD},
+		{u8"D",         JamKey::LowDS},
+		{u8"C",         JamKey::LowE},
+		{u8"V",         JamKey::LowF},
+		{u8"G",         JamKey::LowFS},
+		{u8"B",         JamKey::LowG},
+		{u8"H",         JamKey::LowGS},
+		{u8"N",         JamKey::LowA},
+		{u8"J",         JamKey::LowAS},
+		{u8"M",         JamKey::LowB},
+		{u8",",         JamKey::LowC2},
+		{u8"L",         JamKey::LowCS2},
+		{u8".",         JamKey::LowD2},
 
-		{u8"Q",         JamKey::HIGH_C},
-		{u8"2",         JamKey::HIGH_CS},
-		{u8"W",         JamKey::HIGH_D},
-		{u8"3",         JamKey::HIGH_DS},
-		{u8"E",         JamKey::HIGH_E},
-		{u8"R",         JamKey::HIGH_F},
-		{u8"5",         JamKey::HIGH_FS},
-		{u8"T",         JamKey::HIGH_G},
-		{u8"6",         JamKey::HIGH_GS},
-		{u8"Z",         JamKey::HIGH_A},
-		{u8"7",         JamKey::HIGH_AS},
-		{u8"U",         JamKey::HIGH_B},
-		{u8"I",         JamKey::HIGH_C_H},
-		{u8"9",         JamKey::HIGH_CS_H},
-		{u8"O",         JamKey::HIGH_D_H},
+		{u8"Q",         JamKey::HighC},
+		{u8"2",         JamKey::HighCS},
+		{u8"W",         JamKey::HighD},
+		{u8"3",         JamKey::HighDS},
+		{u8"E",         JamKey::HighE},
+		{u8"R",         JamKey::HighF},
+		{u8"5",         JamKey::HighFS},
+		{u8"T",         JamKey::HighG},
+		{u8"6",         JamKey::HighGS},
+		{u8"Z",         JamKey::HighA},
+		{u8"7",         JamKey::HighAS},
+		{u8"U",         JamKey::HighB},
+		{u8"I",         JamKey::HighC2},
+		{u8"9",         JamKey::HighCS2},
+		{u8"O",         JamKey::HighD2},
 	};
-	const std::map<std::string, JamKey> mappingAZERTY = {
-		{u8"W",         JamKey::LOW_C},
-		{u8"S",         JamKey::LOW_CS},
-		{u8"X",         JamKey::LOW_D},
-		{u8"D",         JamKey::LOW_DS},
-		{u8"C",         JamKey::LOW_E},
-		{u8"V",         JamKey::LOW_F},
-		{u8"G",         JamKey::LOW_FS},
-		{u8"B",         JamKey::LOW_G},
-		{u8"H",         JamKey::LOW_GS},
-		{u8"N",         JamKey::LOW_A},
-		{u8"J",         JamKey::LOW_AS},
-		{u8",",         JamKey::LOW_B},
-		{u8";",         JamKey::LOW_C_H},
-		{u8"L",         JamKey::LOW_CS_H},
-		{u8".",         JamKey::LOW_D_H},
+	const std::unordered_map<std::string, JamKey> mappingAZERTY = {
+		{u8"W",         JamKey::LowC},
+		{u8"S",         JamKey::LowCS},
+		{u8"X",         JamKey::LowD},
+		{u8"D",         JamKey::LowDS},
+		{u8"C",         JamKey::LowE},
+		{u8"V",         JamKey::LowF},
+		{u8"G",         JamKey::LowFS},
+		{u8"B",         JamKey::LowG},
+		{u8"H",         JamKey::LowGS},
+		{u8"N",         JamKey::LowA},
+		{u8"J",         JamKey::LowAS},
+		{u8",",         JamKey::LowB},
+		{u8";",         JamKey::LowC2},
+		{u8"L",         JamKey::LowCS2},
+		{u8".",         JamKey::LowD2},
 
-		{u8"A",         JamKey::HIGH_C},
-		{u8"É",         JamKey::HIGH_CS},   //é - \xc9
-		{u8"Z",         JamKey::HIGH_D},
-		{u8"\"",        JamKey::HIGH_DS},
-		{u8"E",         JamKey::HIGH_E},
-		{u8"R",         JamKey::HIGH_F},
-		{u8"(",         JamKey::HIGH_FS},
-		{u8"T",         JamKey::HIGH_G},
-		{u8"-",         JamKey::HIGH_GS},
-		{u8"Y",         JamKey::HIGH_A},
-		{u8"È",         JamKey::HIGH_AS},   //è - \xc8
-		{u8"U",         JamKey::HIGH_B},
-		{u8"I",         JamKey::HIGH_C_H},
-		{u8"Ç",         JamKey::HIGH_CS_H}, //ç - \xc7
-		{u8"O",         JamKey::HIGH_D_H},
+		{u8"A",         JamKey::HighC},
+		{u8"É",         JamKey::HighCS},   //é - \xc9
+		{u8"Z",         JamKey::HighD},
+		{u8"\"",        JamKey::HighDS},
+		{u8"E",         JamKey::HighE},
+		{u8"R",         JamKey::HighF},
+		{u8"(",         JamKey::HighFS},
+		{u8"T",         JamKey::HighG},
+		{u8"-",         JamKey::HighGS},
+		{u8"Y",         JamKey::HighA},
+		{u8"È",         JamKey::HighAS},   //è - \xc8
+		{u8"U",         JamKey::HighB},
+		{u8"I",         JamKey::HighC2},
+		{u8"Ç",         JamKey::HighCS2}, //ç - \xc7
+		{u8"O",         JamKey::HighD2},
 	};
 	mappingCustom = {};
 	mappingLayouts = {
-		{Custom, mappingCustom},
-		{QWERTY, mappingQWERTY},
-		{QWERTZ, mappingQWERTZ},
-		{AZERTY, mappingAZERTY}
+		{ Custom, mappingCustom },
+		{ QWERTY, mappingQWERTY },
+		{ QWERTZ, mappingQWERTZ },
+		{ AZERTY, mappingAZERTY }
 	};
+
+	// Appearance
+	ptnHdFont_ = u8"";
+	ptnHdFontSize_ = 10;
+	ptnRowFont_ = u8"";
+	ptnRowFontSize_ = 10;
+	odrHdFont_ = u8"";
+	odrHdFontSize_ = 10;
+	odrRowFont_ = u8"";
+	odrRowFontSize_ = 10;
 }
 
 // Internal //
@@ -637,6 +651,16 @@ int Configuration::getMainWindowY() const
 	return mainY_;
 }
 
+void Configuration::setMainWindowVerticalSplit(int y)
+{
+	mainVSplit_ = y;
+}
+
+int Configuration::getMainWindowVerticalSplit() const
+{
+	return mainVSplit_;
+}
+
 void Configuration::setInstrumentFMWindowWidth(int w)
 {
 	instFMW_ = w;
@@ -848,6 +872,26 @@ bool Configuration::getFill00ToEffectValue() const
 	return fill00ToEffectValue_;
 }
 
+void Configuration::setAutosetInstrument(bool enabled)
+{
+	autosetInstrument_ = enabled;
+}
+
+bool Configuration::getAutosetInstrument() const
+{
+	return autosetInstrument_;
+}
+
+void Configuration::setMoveCursorByHorizontalScroll(bool enabled)
+{
+	moveCursorHScroll_ = enabled;
+}
+
+bool Configuration::getMoveCursorByHorizontalScroll() const
+{
+	return moveCursorHScroll_;
+}
+
 // Edit settings
 void Configuration::setPageJumpLength(size_t length)
 {
@@ -930,17 +974,27 @@ Configuration::KeyboardLayout Configuration::getNoteEntryLayout() const
 	return noteEntryLayout_;
 }
 
-void Configuration::setCustomLayoutKeys(std::map<std::string, JamKey> mapping)
+void Configuration::setCustomLayoutKeys(std::unordered_map<std::string, JamKey> mapping)
 {
 	mappingLayouts[KeyboardLayout::Custom] = mapping;
 }
 
-std::map<std::string, JamKey> Configuration::getCustomLayoutKeys() const
+std::unordered_map<std::string, JamKey> Configuration::getCustomLayoutKeys() const
 {
 	return mappingLayouts.at(KeyboardLayout::Custom);
 }
 
 // Sound //
+void Configuration::setSoundAPI(std::string api)
+{
+	sndAPI_ = api;
+}
+
+std::string Configuration::getSoundAPI() const
+{
+	return sndAPI_;
+}
+
 void Configuration::setSoundDevice(std::string device)
 {
 	sndDevice_ = device;
@@ -1044,3 +1098,83 @@ std::vector<FMEnvelopeText> Configuration::getFMEnvelopeTexts() const
 	return fmEnvelopeTexts_;
 }
 
+// Appearrance
+void Configuration::setPatternEditorHeaderFont(std::string font)
+{
+	ptnHdFont_ = font;
+}
+
+std::string Configuration::getPatternEditorHeaderFont() const
+{
+	return ptnHdFont_;
+}
+
+void Configuration::setPatternEditorHeaderFontSize(int size)
+{
+	ptnHdFontSize_ = size;
+}
+
+int Configuration::getPatternEditorHeaderFontSize() const
+{
+	return ptnHdFontSize_;
+}
+
+void Configuration::setPatternEditorRowsFont(std::string font)
+{
+	ptnRowFont_ = font;
+}
+
+std::string Configuration::getPatternEditorRowsFont() const
+{
+	return ptnRowFont_;
+}
+
+void Configuration::setPatternEditorRowsFontSize(int size)
+{
+	ptnRowFontSize_ = size;
+}
+
+int Configuration::getPatternEditorRowsFontSize() const
+{
+	return ptnRowFontSize_;
+}
+
+void Configuration::setOrderListHeaderFont(std::string font)
+{
+	odrHdFont_ = font;
+}
+
+std::string Configuration::getOrderListHeaderFont() const
+{
+	return odrHdFont_;
+}
+
+void Configuration::setOrderListHeaderFontSize(int size)
+{
+	odrHdFontSize_ = size;
+}
+
+int Configuration::getOrderListHeaderFontSize() const
+{
+	return odrHdFontSize_;
+}
+
+void Configuration::setOrderListRowsFont(std::string font)
+{
+	odrRowFont_ = font;
+}
+
+std::string Configuration::getOrderListRowsFont() const
+{
+	return odrRowFont_;
+}
+
+void Configuration::setOrderListRowsFontSize(int size)
+{
+	odrRowFontSize_ = size;
+}
+
+int Configuration::getOrderListRowsFontSize() const
+{
+	return odrRowFontSize_;
+}
