@@ -3,8 +3,9 @@
 #include <cstdint>
 #include <cstddef>
 #include <string>
-#include <map>
+#include <unordered_map>
 #include <vector>
+#include "enum_hash.hpp"
 #include "misc.hpp"
 
 enum class JamKey : int;
@@ -32,6 +33,8 @@ public:
 	int getMainWindowX() const;
 	void setMainWindowY(int y);
 	int getMainWindowY() const;
+	void setMainWindowVerticalSplit(int y);
+	int getMainWindowVerticalSplit() const;
 	void setInstrumentFMWindowWidth(int w);
 	int getInstrumentFMWindowWidth() const;
 	void setInstrumentFMWindowHeight(int h);
@@ -52,6 +55,7 @@ private:
 	int mainW_, mainH_;
 	bool mainMax_;
 	int mainX_, mainY_;
+	int mainVSplit_;
 	int instFMW_, instFMH_;
 	int instSSGW_, instSSGH_;
 	bool followMode_;
@@ -87,15 +91,19 @@ public:
 	bool getShowWaveVisual() const;
 	void setFill00ToEffectValue(bool enabled);
 	bool getFill00ToEffectValue() const;
+	void setAutosetInstrument(bool enabled);
+	bool getAutosetInstrument() const;
+	void setMoveCursorByHorizontalScroll(bool enabled);
+	bool getMoveCursorByHorizontalScroll() const;
 private:
 	bool warpCursor_, warpAcrossOrders_;
 	bool showRowNumHex_, showPrevNextOrders_;
 	bool backupModules_, dontSelectOnDoubleClick_;
 	bool reverseFMVolumeOrder_, moveCursorToRight_;
 	bool retrieveChannelState_, enableTranslation_;
-	bool showFMDetuneSigned_;
-	bool showWaveVisual_;
-	bool fill00ToEffectValue_;
+	bool showFMDetuneSigned_, showWaveVisual_;
+	bool fill00ToEffectValue_, autosetInstrument_;
+	bool moveCursorHScroll_;
 
 	// Edit settings
 public:
@@ -128,13 +136,13 @@ public:
 		QWERTZ,
 		AZERTY
 	};
-	static const std::map<std::string, JamKey> mappingQWERTY, mappingQWERTZ, mappingAZERTY;
-	std::map<std::string, JamKey> mappingCustom;
-	std::map<KeyboardLayout, std::map<std::string, JamKey>> mappingLayouts;
+	static const std::unordered_map<std::string, JamKey> mappingQWERTY, mappingQWERTZ, mappingAZERTY;
+	std::unordered_map<std::string, JamKey> mappingCustom;
+	std::unordered_map<KeyboardLayout, std::unordered_map<std::string, JamKey>> mappingLayouts;
 	void setNoteEntryLayout(KeyboardLayout layout);
 	KeyboardLayout getNoteEntryLayout() const;
-	void setCustomLayoutKeys(std::map<std::string, JamKey> mapping);
-	std::map<std::string, JamKey> getCustomLayoutKeys() const;
+	void setCustomLayoutKeys(std::unordered_map<std::string, JamKey> mapping);
+	std::unordered_map<std::string, JamKey> getCustomLayoutKeys() const;
 
 private:
 	std::string keyOffKey_, octUpKey_, octDownKey_, echoKey_;
@@ -142,6 +150,8 @@ private:
 
 	// Sound //
 public:
+	void setSoundAPI(std::string api);
+	std::string getSoundAPI() const;
 	void setSoundDevice(std::string device);
 	std::string getSoundDevice() const;
 	void setUseSCCI(bool enabled);
@@ -153,7 +163,7 @@ public:
 	void setBufferLength(size_t length);
 	size_t getBufferLength() const;
 private:
-	std::string sndDevice_;
+	std::string sndAPI_, sndDevice_;
 	bool useSCCI_;
 	int emulator_;
 	uint32_t sampleRate_;
@@ -183,6 +193,28 @@ public:
 	void setFMEnvelopeTexts(std::vector<FMEnvelopeText> texts);
 	std::vector<FMEnvelopeText> getFMEnvelopeTexts() const;
 
+	// Appearance //
+public:
+	void setPatternEditorHeaderFont(std::string font);
+	std::string getPatternEditorHeaderFont() const;
+	void setPatternEditorHeaderFontSize(int size);
+	int getPatternEditorHeaderFontSize() const;
+	void setPatternEditorRowsFont(std::string font);
+	std::string getPatternEditorRowsFont() const;
+	void setPatternEditorRowsFontSize(int size);
+	int getPatternEditorRowsFontSize() const;
+	void setOrderListHeaderFont(std::string font);
+	std::string getOrderListHeaderFont() const;
+	void setOrderListHeaderFontSize(int size);
+	int getOrderListHeaderFontSize() const;
+	void setOrderListRowsFont(std::string font);
+	std::string getOrderListRowsFont() const;
+	void setOrderListRowsFontSize(int size);
+	int getOrderListRowsFontSize() const;
+private:
+	std::string ptnHdFont_, ptnRowFont_, odrHdFont_, odrRowFont_;
+	int ptnHdFontSize_, ptnRowFontSize_, odrHdFontSize_, odrRowFontSize_;
+
 private:
 	std::vector<FMEnvelopeText> fmEnvelopeTexts_;
 };
