@@ -1,7 +1,9 @@
 #include "configuration_handler.hpp"
 #include <vector>
+#include <unordered_map>
 #include <QSettings>
 #include "jam_manager.hpp"
+#include"enum_hash.hpp"
 
 // config path (*nix): ~/.config/<organization>/<application>.ini
 const QString ConfigurationHandler::organization = "BambooTracker";
@@ -22,6 +24,7 @@ bool ConfigurationHandler::saveConfiguration(std::weak_ptr<Configuration> config
 		settings.setValue("mainWindowMaximized",       configLocked->getMainWindowMaximized());
 		settings.setValue("mainWindowX",               configLocked->getMainWindowX());
 		settings.setValue("mainWindowY",               configLocked->getMainWindowY());
+		settings.setValue("mainWindowVerticalSplit",	configLocked->getMainWindowVerticalSplit());
 		settings.setValue("instrumentFMWindowWidth",   configLocked->getInstrumentFMWindowWidth());
 		settings.setValue("instrumentFMWindowHeight",  configLocked->getInstrumentFMWindowHeight());
 		settings.setValue("instrumentSSGWindowWidth",  configLocked->getInstrumentSSGWindowWidth());
@@ -48,6 +51,8 @@ bool ConfigurationHandler::saveConfiguration(std::weak_ptr<Configuration> config
 		settings.setValue("showFMDetuneAsSigned",	configLocked->getShowFMDetuneAsSigned());
 		settings.setValue("showWaveVisual",			configLocked->getShowWaveVisual());
 		settings.setValue("fill00ToEffectValue",	configLocked->getFill00ToEffectValue());
+		settings.setValue("autosetInstrument",		configLocked->getAutosetInstrument());
+		settings.setValue("moveCursorByHScroll",	configLocked->getMoveCursorByHorizontalScroll());
 		settings.endGroup();
 
 		// Edit settings
@@ -72,41 +77,41 @@ bool ConfigurationHandler::saveConfiguration(std::weak_ptr<Configuration> config
 						  QString::fromUtf8(configLocked->getEchoBufferKey().c_str(),
 											static_cast<int>(configLocked->getEchoBufferKey().length())));
 		settings.setValue("noteEntryLayout",	static_cast<int>(configLocked->getNoteEntryLayout()));
-		std::map<std::string, JamKey> customLayoutMapping = configLocked->getCustomLayoutKeys();
-		const std::map<JamKey, std::string> keyToNameMapping = {
-			{JamKey::LOW_C,     "lowC"},
-			{JamKey::LOW_CS,    "lowCS"},
-			{JamKey::LOW_D,     "lowD"},
-			{JamKey::LOW_DS,    "lowDS"},
-			{JamKey::LOW_E,     "lowE"},
-			{JamKey::LOW_F,     "lowF"},
-			{JamKey::LOW_FS,    "lowFS"},
-			{JamKey::LOW_G,     "lowG"},
-			{JamKey::LOW_GS,    "lowGS"},
-			{JamKey::LOW_A,     "lowA"},
-			{JamKey::LOW_AS,    "lowAS"},
-			{JamKey::LOW_B,     "lowB"},
-			{JamKey::LOW_C_H,   "lowHighC"},
-			{JamKey::LOW_CS_H,  "lowHighCS"},
-			{JamKey::LOW_D_H,   "lowHighD"},
+		std::unordered_map<std::string, JamKey> customLayoutMapping = configLocked->getCustomLayoutKeys();
+		const std::unordered_map<JamKey, std::string> keyToNameMapping = {
+			{JamKey::LowC,     "lowC"},
+			{JamKey::LowCS,    "lowCS"},
+			{JamKey::LowD,     "lowD"},
+			{JamKey::LowDS,    "lowDS"},
+			{JamKey::LowE,     "lowE"},
+			{JamKey::LowF,     "lowF"},
+			{JamKey::LowFS,    "lowFS"},
+			{JamKey::LowG,     "lowG"},
+			{JamKey::LowGS,    "lowGS"},
+			{JamKey::LowA,     "lowA"},
+			{JamKey::LowAS,    "lowAS"},
+			{JamKey::LowB,     "lowB"},
+			{JamKey::LowC2,   "lowHighC"},
+			{JamKey::LowCS2,  "lowHighCS"},
+			{JamKey::LowD2,   "lowHighD"},
 
-			{JamKey::HIGH_C,    "highC"},
-			{JamKey::HIGH_CS,   "highCS"},
-			{JamKey::HIGH_D,    "highD"},
-			{JamKey::HIGH_DS,   "highDS"},
-			{JamKey::HIGH_E,    "highE"},
-			{JamKey::HIGH_F,    "highF"},
-			{JamKey::HIGH_FS,   "highFS"},
-			{JamKey::HIGH_G,    "highG"},
-			{JamKey::HIGH_GS,   "highGS"},
-			{JamKey::HIGH_A,    "highA"},
-			{JamKey::HIGH_AS,   "highAS"},
-			{JamKey::HIGH_B,    "highB"},
-			{JamKey::HIGH_C_H,  "highHighC"},
-			{JamKey::HIGH_CS_H, "highHighCS"},
-			{JamKey::HIGH_D_H,  "highHighD"}
+			{JamKey::HighC,    "highC"},
+			{JamKey::HighCS,   "highCS"},
+			{JamKey::HighD,    "highD"},
+			{JamKey::HighDS,   "highDS"},
+			{JamKey::HighE,    "highE"},
+			{JamKey::HighF,    "highF"},
+			{JamKey::HighFS,   "highFS"},
+			{JamKey::HighG,    "highG"},
+			{JamKey::HighGS,   "highGS"},
+			{JamKey::HighA,    "highA"},
+			{JamKey::HighAS,   "highAS"},
+			{JamKey::HighB,    "highB"},
+			{JamKey::HighC2,  "highHighC"},
+			{JamKey::HighCS2, "highHighCS"},
+			{JamKey::HighD2,  "highHighD"}
 		};
-		std::map<std::string, JamKey>::const_iterator customLayoutMappingIterator = customLayoutMapping.begin();
+		std::unordered_map<std::string, JamKey>::const_iterator customLayoutMappingIterator = customLayoutMapping.begin();
 		while (customLayoutMappingIterator != customLayoutMapping.end()) {
 			settings.setValue(QString::fromStdString("customLayout_" + keyToNameMapping.at(customLayoutMappingIterator->second)),
 							  QString::fromUtf8(customLayoutMappingIterator->first.c_str(), static_cast<int> (customLayoutMappingIterator->first.length())));
@@ -116,6 +121,9 @@ bool ConfigurationHandler::saveConfiguration(std::weak_ptr<Configuration> config
 
 		// Sound //
 		settings.beginGroup("Sound");
+		settings.setValue("soundAPI",
+						  QString::fromUtf8(configLocked->getSoundAPI().c_str(),
+											static_cast<int>(configLocked->getSoundAPI().length())));
 		settings.setValue("soundDevice",
 						  QString::fromUtf8(configLocked->getSoundDevice().c_str(),
 											static_cast<int>(configLocked->getSoundDevice().length())));
@@ -144,16 +152,25 @@ bool ConfigurationHandler::saveConfiguration(std::weak_ptr<Configuration> config
 		for (auto texts : config.lock()->getFMEnvelopeTexts()) {
 			settings.setArrayIndex(n++);
 			settings.setValue("type", QString::fromUtf8(texts.name.c_str(), static_cast<int>(texts.name.length())));
-			QString data;
-			for (auto type : texts.texts) {
-				data += QString(",%1").arg(static_cast<int>(type));
-			}
-			if (!data.isEmpty()) data.remove(0, 1);
-			settings.setValue("order", data);
+			QStringList typeList;
+			std::transform(texts.texts.begin(), texts.texts.end(), std::back_inserter(typeList),
+						   [](FMEnvelopeTextType type) { return QString::number(static_cast<int>(type)); });
+			settings.setValue("order", typeList.join(","));
 		}
 		settings.endArray();
 		settings.endGroup();
 
+		settings.beginGroup("Appearance");
+		settings.setValue("patternEditorHeaderFont", QString::fromStdString(configLocked->getPatternEditorHeaderFont()));
+		settings.setValue("patternEditorHeaderFontSize", configLocked->getPatternEditorHeaderFontSize());
+		settings.setValue("patternEditorRowsFont", QString::fromStdString(configLocked->getPatternEditorRowsFont()));
+		settings.setValue("patternEditorRowsFontSize", configLocked->getPatternEditorRowsFontSize());
+		settings.setValue("orderListHeaderFont", QString::fromStdString(configLocked->getOrderListHeaderFont()));
+		settings.setValue("orderListHeaderFontSize", configLocked->getOrderListHeaderFontSize());
+		settings.setValue("orderListRowsFont", QString::fromStdString(configLocked->getOrderListRowsFont()));
+		settings.setValue("orderListRowsFontSize", configLocked->getOrderListRowsFontSize());
+		settings.endGroup();
+
 		return true;
 	} catch (...) {
 		return false;
@@ -173,6 +190,7 @@ bool ConfigurationHandler::loadConfiguration(std::weak_ptr<Configuration> config
 		configLocked->setMainWindowMaximized(settings.value("mainWindowMaximized", configLocked->getMainWindowMaximized()).toBool());
 		configLocked->setMainWindowX(settings.value("mainWindowX", configLocked->getMainWindowX()).toInt());
 		configLocked->setMainWindowY(settings.value("mainWindowY", configLocked->getMainWindowY()).toInt());
+		configLocked->setMainWindowVerticalSplit(settings.value("mainWindowVerticalSplit", configLocked->getMainWindowVerticalSplit()).toInt());
 		configLocked->setInstrumentFMWindowWidth(settings.value("instrumentFMWindowWidth", configLocked->getInstrumentFMWindowWidth()).toInt());
 		configLocked->setInstrumentFMWindowHeight(settings.value("instrumentFMWindowHeight", configLocked->getInstrumentFMWindowHeight()).toInt());
 		configLocked->setInstrumentSSGWindowWidth(settings.value("instrumentSSGWindowWidth", configLocked->getInstrumentSSGWindowWidth()).toInt());
@@ -199,6 +217,8 @@ bool ConfigurationHandler::loadConfiguration(std::weak_ptr<Configuration> config
 		configLocked->setShowFMDetuneAsSigned(settings.value("showFMDetuneAsSigned", configLocked->getShowFMDetuneAsSigned()).toBool());
 		configLocked->setShowWaveVisual(settings.value("showWaveVisual", configLocked->getShowWaveVisual()).toBool());
 		configLocked->setFill00ToEffectValue(settings.value("fill00ToEffectValue", configLocked->getFill00ToEffectValue()).toBool());
+		configLocked->setAutosetInstrument(settings.value("autosetInstrument", configLocked->getAutosetInstrument()).toBool());
+		configLocked->setMoveCursorByHorizontalScroll(settings.value("moveCursorByHScroll", configLocked->getMoveCursorByHorizontalScroll()).toBool());
 		settings.endGroup();
 
 		// Edit settings
@@ -221,41 +241,41 @@ bool ConfigurationHandler::loadConfiguration(std::weak_ptr<Configuration> config
 		configLocked->setNoteEntryLayout(static_cast<Configuration::KeyboardLayout>(
 											 settings.value("noteEntryLayout",
 															static_cast<int>(configLocked->getNoteEntryLayout())).toInt()));
-		std::map<std::string, JamKey> customLayoutNewKeys = {};
-		const std::map<std::string, JamKey> nameToKeyMapping = {
-			{"lowC",       JamKey::LOW_C},
-			{"lowCS",      JamKey::LOW_CS},
-			{"lowD",       JamKey::LOW_D},
-			{"lowDS",      JamKey::LOW_DS},
-			{"lowE",       JamKey::LOW_E},
-			{"lowF",       JamKey::LOW_F},
-			{"lowFS",      JamKey::LOW_FS},
-			{"lowG",       JamKey::LOW_G},
-			{"lowGS",      JamKey::LOW_GS},
-			{"lowA",       JamKey::LOW_A},
-			{"lowAS",      JamKey::LOW_AS},
-			{"lowB",       JamKey::LOW_B},
-			{"lowHighC",   JamKey::LOW_C_H},
-			{"lowHighCS",  JamKey::LOW_CS_H},
-			{"lowHighD",   JamKey::LOW_D_H},
+		std::unordered_map<std::string, JamKey> customLayoutNewKeys;
+		const std::unordered_map<std::string, JamKey> nameToKeyMapping = {
+			{"lowC",       JamKey::LowC},
+			{"lowCS",      JamKey::LowCS},
+			{"lowD",       JamKey::LowD},
+			{"lowDS",      JamKey::LowDS},
+			{"lowE",       JamKey::LowE},
+			{"lowF",       JamKey::LowF},
+			{"lowFS",      JamKey::LowFS},
+			{"lowG",       JamKey::LowG},
+			{"lowGS",      JamKey::LowGS},
+			{"lowA",       JamKey::LowA},
+			{"lowAS",      JamKey::LowAS},
+			{"lowB",       JamKey::LowB},
+			{"lowHighC",   JamKey::LowC2},
+			{"lowHighCS",  JamKey::LowCS2},
+			{"lowHighD",   JamKey::LowD2},
 
-			{"highC",      JamKey::HIGH_C},
-			{"highCS",     JamKey::HIGH_CS},
-			{"highD",      JamKey::HIGH_D},
-			{"highDS",     JamKey::HIGH_DS},
-			{"highE",      JamKey::HIGH_E},
-			{"highF",      JamKey::HIGH_F},
-			{"highFS",     JamKey::HIGH_FS},
-			{"highG",      JamKey::HIGH_G},
-			{"highGS",     JamKey::HIGH_GS},
-			{"highA",      JamKey::HIGH_A},
-			{"highAS",     JamKey::HIGH_AS},
-			{"highB",      JamKey::HIGH_B},
-			{"highHighC",  JamKey::HIGH_C_H},
-			{"highHighCS", JamKey::HIGH_CS_H},
-			{"highHighD",  JamKey::HIGH_D_H}
+			{"highC",      JamKey::HighC},
+			{"highCS",     JamKey::HighCS},
+			{"highD",      JamKey::HighD},
+			{"highDS",     JamKey::HighDS},
+			{"highE",      JamKey::HighE},
+			{"highF",      JamKey::HighF},
+			{"highFS",     JamKey::HighFS},
+			{"highG",      JamKey::HighG},
+			{"highGS",     JamKey::HighGS},
+			{"highA",      JamKey::HighA},
+			{"highAS",     JamKey::HighAS},
+			{"highB",      JamKey::HighB},
+			{"highHighC",  JamKey::HighC2},
+			{"highHighCS", JamKey::HighCS2},
+			{"highHighD",  JamKey::HighD2}
 		};
-		std::map<std::string, JamKey>::const_iterator nameToKeyMappingIterator = nameToKeyMapping.begin();
+		std::unordered_map<std::string, JamKey>::const_iterator nameToKeyMappingIterator = nameToKeyMapping.begin();
 		while (nameToKeyMappingIterator != nameToKeyMapping.end()) {
 			JamKey currentlyWantedJamKey = nameToKeyMappingIterator->second;
 			customLayoutNewKeys[
@@ -272,6 +292,7 @@ bool ConfigurationHandler::loadConfiguration(std::weak_ptr<Configuration> config
 
 		// Sound //
 		settings.beginGroup("Sound");
+		configLocked->setSoundAPI(settings.value("soundAPI", QString::fromStdString(configLocked->getSoundAPI())).toString().toUtf8().toStdString());
 		configLocked->setSoundDevice(settings.value("soundDevice", QString::fromStdString(configLocked->getSoundDevice())).toString().toUtf8().toStdString());
 		configLocked->setUseSCCI(settings.value("useSCCI", configLocked->getUseSCCI()).toBool());
 		configLocked->setEmulator(settings.value("emulator", configLocked->getEmulator()).toInt());
@@ -312,6 +333,18 @@ bool ConfigurationHandler::loadConfiguration(std::weak_ptr<Configuration> config
 		settings.endArray();
 		settings.endGroup();
 
+		// Appearance
+		settings.beginGroup("Appearance");
+		configLocked->setPatternEditorHeaderFont(settings.value("patternEditorHeaderFont", QString::fromStdString(configLocked->getPatternEditorHeaderFont())).toString().toStdString());
+		configLocked->setPatternEditorHeaderFontSize(settings.value("patternEditorHeaderFontSize", configLocked->getPatternEditorHeaderFontSize()).toInt());
+		configLocked->setPatternEditorRowsFont(settings.value("patternEditorRowsFont", QString::fromStdString(configLocked->getPatternEditorRowsFont())).toString().toStdString());
+		configLocked->setPatternEditorRowsFontSize(settings.value("patternEditorRowsFontSize", configLocked->getPatternEditorRowsFontSize()).toInt());
+		configLocked->setOrderListHeaderFont(settings.value("orderListHeaderFont", QString::fromStdString(configLocked->getOrderListHeaderFont())).toString().toStdString());
+		configLocked->setOrderListHeaderFontSize(settings.value("orderListHeaderFontSize", configLocked->getOrderListHeaderFontSize()).toInt());
+		configLocked->setOrderListRowsFont(settings.value("orderListRowsFont", QString::fromStdString(configLocked->getOrderListRowsFont())).toString().toStdString());
+		configLocked->setOrderListRowsFontSize(settings.value("orderListRowsFontSize", configLocked->getOrderListRowsFontSize()).toInt());
+		settings.endGroup();
+
 		return true;
 	} catch (...) {
 		return false;
