@@ -13,6 +13,7 @@
 #include <QStringList>
 #include <QMessageBox>
 #include "gui/event_guard.hpp"
+#include "gui/jam_layout.hpp"
 #include "misc.hpp"
 
 InstrumentEditorFMForm::InstrumentEditorFMForm(int num, QWidget *parent) :
@@ -219,7 +220,7 @@ InstrumentEditorFMForm::InstrumentEditorFMForm(int num, QWidget *parent) :
 	/******************** LFO editor ********************/
 	ui->lfoGroupBox->setContextMenuPolicy(Qt::CustomContextMenu);
 
-	ui->lfoFreqSlider->setText("Freq");
+	ui->lfoFreqSlider->setText(tr("Freq"));
 	ui->lfoFreqSlider->setMaximum(7);
 	QObject::connect(ui->lfoFreqSlider, &LabeledVerticalSlider::valueChanged,
 					 this, [&](int v) {
@@ -292,7 +293,7 @@ InstrumentEditorFMForm::InstrumentEditorFMForm(int num, QWidget *parent) :
 	QObject::connect(ui->lfoStartSpinBox, static_cast<void(QSpinBox::*)(int)>(&QSpinBox::valueChanged),
 					 this, [&](int v) {
 		if (!isIgnoreEvent_) {
-			bt_.lock()->setLFOFMParameter(ui->lfoNumSpinBox->value(), FMLFOParameter::COUNT, v);
+			bt_.lock()->setLFOFMParameter(ui->lfoNumSpinBox->value(), FMLFOParameter::Count, v);
 			emit lfoParameterChanged(ui->lfoNumSpinBox->value(), instNum_);
 			emit modified();
 		}
@@ -396,28 +397,15 @@ InstrumentEditorFMForm::InstrumentEditorFMForm(int num, QWidget *parent) :
 					 this, &InstrumentEditorFMForm::onOperatorSequenceTypeChanged);
 
 	//========== Arpeggio ==========//
-	ui->arpOpComboBox->addItem("All", static_cast<int>(FMOperatorType::All));
+	ui->arpOpComboBox->addItem(tr("All"), static_cast<int>(FMOperatorType::All));
 	ui->arpOpComboBox->addItem("Op1", static_cast<int>(FMOperatorType::Op1));
 	ui->arpOpComboBox->addItem("Op2", static_cast<int>(FMOperatorType::Op2));
 	ui->arpOpComboBox->addItem("Op3", static_cast<int>(FMOperatorType::Op3));
 	ui->arpOpComboBox->addItem("Op4", static_cast<int>(FMOperatorType::Op4));
 
-	ui->arpEditor->setMaximumDisplayedRowCount(15);
-	ui->arpEditor->setDefaultRow(48);
-	ui->arpEditor->setLabelDiaplayMode(true);
-	for (int i = 0; i < 96; ++i) {
-		int d = i - 48;
-		auto text = QString::number(d);
-		if (d > 0) text = "+" + text;
-		ui->arpEditor->AddRow(text, false);
-	}
-	ui->arpEditor->autoFitLabelWidth();
-	ui->arpEditor->setUpperRow(55);
-	ui->arpEditor->setMMLDisplay0As(-48);
-
-	ui->arpTypeComboBox->addItem(tr("Absolute"), 0);
-	ui->arpTypeComboBox->addItem(tr("Fix"), 1);
-	ui->arpTypeComboBox->addItem(tr("Relative"), 2);
+	ui->arpTypeComboBox->addItem(tr("Absolute"), VisualizedInstrumentMacroEditor::SequenceType::Absolute);
+	ui->arpTypeComboBox->addItem(tr("Fixed"), VisualizedInstrumentMacroEditor::SequenceType::Fixed);
+	ui->arpTypeComboBox->addItem(tr("Relative"), VisualizedInstrumentMacroEditor::SequenceType::Relative);
 
 	QObject::connect(ui->arpEditor, &VisualizedInstrumentMacroEditor::sequenceCommandAdded,
 					 this, [&](int row, int col) {
@@ -471,7 +459,7 @@ InstrumentEditorFMForm::InstrumentEditorFMForm(int num, QWidget *parent) :
 					 this, &InstrumentEditorFMForm::onArpeggioOperatorChanged);
 
 	//========== Pitch ==========//
-	ui->ptOpComboBox->addItem("All", static_cast<int>(FMOperatorType::All));
+	ui->ptOpComboBox->addItem(tr("All"), static_cast<int>(FMOperatorType::All));
 	ui->ptOpComboBox->addItem("Op1", static_cast<int>(FMOperatorType::Op1));
 	ui->ptOpComboBox->addItem("Op2", static_cast<int>(FMOperatorType::Op2));
 	ui->ptOpComboBox->addItem("Op3", static_cast<int>(FMOperatorType::Op3));
@@ -481,17 +469,14 @@ InstrumentEditorFMForm::InstrumentEditorFMForm(int num, QWidget *parent) :
 	ui->ptEditor->setDefaultRow(127);
 	ui->ptEditor->setLabelDiaplayMode(true);
 	for (int i = 0; i < 255; ++i) {
-		int d = i - 127;
-		auto text = QString::number(d);
-		if (d > 0) text = "+" + text;
-		ui->ptEditor->AddRow(text, false);
+		ui->ptEditor->AddRow(QString::asprintf("%+d", i - 127), false);
 	}
 	ui->ptEditor->autoFitLabelWidth();
 	ui->ptEditor->setUpperRow(134);
 	ui->ptEditor->setMMLDisplay0As(-127);
 
-	ui->ptTypeComboBox->addItem("Absolute", 0);
-	ui->ptTypeComboBox->addItem("Relative", 2);
+	ui->ptTypeComboBox->addItem(tr("Absolute"), VisualizedInstrumentMacroEditor::SequenceType::Absolute);
+	ui->ptTypeComboBox->addItem(tr("Relative"), VisualizedInstrumentMacroEditor::SequenceType::Relative);
 
 	QObject::connect(ui->ptEditor, &VisualizedInstrumentMacroEditor::sequenceCommandAdded,
 					 this, [&](int row, int col) {
@@ -636,17 +621,49 @@ void InstrumentEditorFMForm::setColorPalette(std::shared_ptr<ColorPalette> palet
 	ui->ptEditor->setColorPalette(palette);
 }
 
+SequenceType InstrumentEditorFMForm::convertSequenceTypeForData(VisualizedInstrumentMacroEditor::SequenceType type)
+{
+	switch (type) {
+	case VisualizedInstrumentMacroEditor::SequenceType::NoType:
+		return SequenceType::NO_SEQUENCE_TYPE;
+	case VisualizedInstrumentMacroEditor::SequenceType::Fixed:
+		return SequenceType::FIXED_SEQUENCE;
+	case VisualizedInstrumentMacroEditor::SequenceType::Absolute:
+		return SequenceType::ABSOLUTE_SEQUENCE;
+	case VisualizedInstrumentMacroEditor::SequenceType::Relative:
+		return SequenceType::RELATIVE_SEQUENCE;
+	default:
+		throw std::invalid_argument("Unexpected SequenceType.");
+	}
+}
+
+VisualizedInstrumentMacroEditor::SequenceType InstrumentEditorFMForm::convertSequenceTypeForUI(SequenceType type)
+{
+	switch (type) {
+	case SequenceType::NO_SEQUENCE_TYPE:
+		return VisualizedInstrumentMacroEditor::SequenceType::NoType;
+	case SequenceType::FIXED_SEQUENCE:
+		return VisualizedInstrumentMacroEditor::SequenceType::Fixed;
+	case SequenceType::ABSOLUTE_SEQUENCE:
+		return VisualizedInstrumentMacroEditor::SequenceType::Absolute;
+	case SequenceType::RELATIVE_SEQUENCE:
+		return VisualizedInstrumentMacroEditor::SequenceType::Relative;
+	default:
+		throw std::invalid_argument("Unexpected SequenceType.");
+	}
+}
+
 ReleaseType InstrumentEditorFMForm::convertReleaseTypeForData(VisualizedInstrumentMacroEditor::ReleaseType type)
 {
 	switch (type) {
 	case VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE:
-		return ReleaseType::NO_RELEASE;
-	case VisualizedInstrumentMacroEditor::ReleaseType::FIX:
-		return ReleaseType::FIX;
+		return ReleaseType::NoRelease;
+	case VisualizedInstrumentMacroEditor::ReleaseType::FIXED:
+		return ReleaseType::FixedRelease;
 	case VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE:
-		return ReleaseType::ABSOLUTE;
+		return ReleaseType::AbsoluteRelease;
 	case VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE:
-		return ReleaseType::RELATIVE;
+		return ReleaseType::RelativeRelease;
 	default:
 		throw std::invalid_argument("Unexpected ReleaseType.");
 	}
@@ -655,13 +672,13 @@ ReleaseType InstrumentEditorFMForm::convertReleaseTypeForData(VisualizedInstrume
 VisualizedInstrumentMacroEditor::ReleaseType InstrumentEditorFMForm::convertReleaseTypeForUI(ReleaseType type)
 {
 	switch (type) {
-	case ReleaseType::NO_RELEASE:
+	case ReleaseType::NoRelease:
 		return VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE;
-	case ReleaseType::FIX:
-		return VisualizedInstrumentMacroEditor::ReleaseType::FIX;
-	case ReleaseType::ABSOLUTE:
+	case ReleaseType::FixedRelease:
+		return VisualizedInstrumentMacroEditor::ReleaseType::FIXED;
+	case ReleaseType::AbsoluteRelease:
 		return VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE;
-	case ReleaseType::RELATIVE:
+	case ReleaseType::RelativeRelease:
 		return VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE;
 	default:
 		throw std::invalid_argument("Unexpected ReleaseType.");
@@ -718,41 +735,11 @@ void InstrumentEditorFMForm::keyPressEvent(QKeyEvent *event)
 	default:
 		if (!event->isAutoRepeat()) {
 			// Musical keyboard
-			switch (event->key()) {
-			case Qt::Key_Z:
-			case Qt::Key_S:
-			case Qt::Key_X:
-			case Qt::Key_D:
-			case Qt::Key_C:
-			case Qt::Key_V:
-			case Qt::Key_G:
-			case Qt::Key_B:
-			case Qt::Key_H:
-			case Qt::Key_N:
-			case Qt::Key_J:
-			case Qt::Key_M:
-			case Qt::Key_Comma:
-			case Qt::Key_L:
-			case Qt::Key_Period:
-			case Qt::Key_Q:
-			case Qt::Key_2:
-			case Qt::Key_W:
-			case Qt::Key_3:
-			case Qt::Key_E:
-			case Qt::Key_R:
-			case Qt::Key_5:
-			case Qt::Key_T:
-			case Qt::Key_6:
-			case Qt::Key_Y:
-			case Qt::Key_7:
-			case Qt::Key_U:
-			case Qt::Key_I:
-			case Qt::Key_9:
-			case Qt::Key_O:
-				emit jamKeyOnEvent(event);
-				break;
-			default: break;
-			}
+			Qt::Key qtKey = static_cast<Qt::Key>(event->key());
+			try {
+				JamKey jk = getJamKeyFromLayoutMapping(qtKey, config_);
+				emit jamKeyOnEvent(jk);
+			} catch (std::invalid_argument&) {}
 		}
 		break;
 	}
@@ -763,41 +750,11 @@ void InstrumentEditorFMForm::keyReleaseEvent(QKeyEvent *event)
 {
 	// For jam key off
 	if (!event->isAutoRepeat()) {
-		switch (event->key()) {
-		case Qt::Key_Z:
-		case Qt::Key_S:
-		case Qt::Key_X:
-		case Qt::Key_D:
-		case Qt::Key_C:
-		case Qt::Key_V:
-		case Qt::Key_G:
-		case Qt::Key_B:
-		case Qt::Key_H:
-		case Qt::Key_N:
-		case Qt::Key_J:
-		case Qt::Key_M:
-		case Qt::Key_Comma:
-		case Qt::Key_L:
-		case Qt::Key_Period:
-		case Qt::Key_Q:
-		case Qt::Key_2:
-		case Qt::Key_W:
-		case Qt::Key_3:
-		case Qt::Key_E:
-		case Qt::Key_R:
-		case Qt::Key_5:
-		case Qt::Key_T:
-		case Qt::Key_6:
-		case Qt::Key_Y:
-		case Qt::Key_7:
-		case Qt::Key_U:
-		case Qt::Key_I:
-		case Qt::Key_9:
-		case Qt::Key_O:
-			emit jamKeyOffEvent(event);
-			break;
-		default: break;
-		}
+		Qt::Key qtKey = static_cast<Qt::Key>(event->key());
+		try {
+			JamKey jk = getJamKeyFromLayoutMapping(qtKey, config_);
+			emit jamKeyOffEvent(jk);
+		} catch (std::invalid_argument&) {}
 	}
 }
 
@@ -1359,13 +1316,12 @@ void InstrumentEditorFMForm::onEnvelopeParameterChanged(int envNum)
 void InstrumentEditorFMForm::onEnvelopeNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getEnvelopeFMUsers(ui->envNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-	ui->envUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->envUsersLineEdit->setText(l.join((",")));
 }
 
 //========== LFO ==========//
@@ -1384,7 +1340,7 @@ void InstrumentEditorFMForm::setInstrumentLFOParameters()
 	ui->amOp2CheckBox->setChecked(instFM->getLFOParameter(FMLFOParameter::AM2));
 	ui->amOp3CheckBox->setChecked(instFM->getLFOParameter(FMLFOParameter::AM3));
 	ui->amOp4CheckBox->setChecked(instFM->getLFOParameter(FMLFOParameter::AM4));
-	ui->lfoStartSpinBox->setValue(instFM->getLFOParameter(FMLFOParameter::COUNT));
+	ui->lfoStartSpinBox->setValue(instFM->getLFOParameter(FMLFOParameter::Count));
 	if (instFM->getLFOEnabled()) {
 		ui->lfoGroupBox->setChecked(true);
 		onLFONumberChanged();
@@ -1438,13 +1394,12 @@ void InstrumentEditorFMForm::onLFOParameterChanged(int lfoNum)
 void InstrumentEditorFMForm::onLFONumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getLFOFMUsers(ui->lfoNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-	ui->lfoUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->lfoUsersLineEdit->setText(l.join(","));
 }
 
 void InstrumentEditorFMForm::on_lfoGroupBox_customContextMenuRequested(const QPoint &pos)
@@ -1617,14 +1572,12 @@ void InstrumentEditorFMForm::setOperatorSequenceEditor()
 void InstrumentEditorFMForm::onOperatorSequenceNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getOperatorSequenceFMUsers(getOperatorSequenceParameter(), ui->opSeqNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->opSeqUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->opSeqUsersLineEdit->setText(l.join(","));
 }
 
 void InstrumentEditorFMForm::onOperatorSequenceParameterChanged(FMEnvelopeParameter param, int tnNum)
@@ -1691,7 +1644,13 @@ void InstrumentEditorFMForm::setInstrumentArpeggioParameters()
 	}
 	ui->arpEditor->setRelease(convertReleaseTypeForUI(instFM->getArpeggioRelease(param).type),
 							  instFM->getArpeggioRelease(param).begin);
-	ui->arpTypeComboBox->setCurrentIndex(instFM->getArpeggioType(param));
+	for (int i = 0; i < ui->arpTypeComboBox->count(); ++i) {
+		if (instFM->getArpeggioType(param) == convertSequenceTypeForData(
+					static_cast<VisualizedInstrumentMacroEditor::SequenceType>(ui->arpTypeComboBox->itemData(i).toInt()))) {
+			ui->arpTypeComboBox->setCurrentIndex(i);
+			break;
+		}
+	}
 	if (instFM->getArpeggioEnabled(param)) {
 		ui->arpEditGroupBox->setChecked(true);
 		onArpeggioNumberChanged();
@@ -1705,14 +1664,12 @@ void InstrumentEditorFMForm::setInstrumentArpeggioParameters()
 void InstrumentEditorFMForm::onArpeggioNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getArpeggioFMUsers(ui->arpNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->arpUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->arpUsersLineEdit->setText(l.join(","));
 }
 
 void InstrumentEditorFMForm::onArpeggioParameterChanged(int tnNum)
@@ -1734,30 +1691,15 @@ void InstrumentEditorFMForm::onArpeggioTypeChanged(int index)
 {
 	Q_UNUSED(index)
 
+	auto type = static_cast<VisualizedInstrumentMacroEditor::SequenceType>(
+					ui->arpTypeComboBox->currentData(Qt::UserRole).toInt());
 	if (!isIgnoreEvent_) {
-		bt_.lock()->setArpeggioFMType(ui->arpNumSpinBox->value(),
-									  ui->arpTypeComboBox->currentData(Qt::UserRole).toInt());
+		bt_.lock()->setArpeggioFMType(ui->arpNumSpinBox->value(), convertSequenceTypeForData(type));
 		emit arpeggioParameterChanged(ui->arpNumSpinBox->value(), instNum_);
 		emit modified();
 	}
 
-	// Update labels
-	if (index == 1) {
-		QString tn[] = { "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B" };
-		for (int i = 0; i < 8; ++i) {
-			for (int j = 0; j < 12; ++j) {
-				ui->arpEditor->setLabel(i * 12 + j, tn[j] + QString::number(i));
-			}
-		}
-	}
-	else {
-		for (int i = 0; i < 96; ++i) {
-			int d = i - 48;
-			auto text = QString::number(d);
-			if (d > 0) text = "+" + text;
-			ui->arpEditor->setLabel(i, text);
-		}
-	}
+	ui->arpEditor->setSequenceType(type);
 }
 
 void InstrumentEditorFMForm::on_arpEditGroupBox_toggled(bool arg1)
@@ -1810,7 +1752,8 @@ void InstrumentEditorFMForm::setInstrumentPitchParameters()
 	ui->ptEditor->setRelease(convertReleaseTypeForUI(instFM->getPitchRelease(param).type),
 							 instFM->getPitchRelease(param).begin);
 	for (int i = 0; i < ui->ptTypeComboBox->count(); ++i) {
-		if (ui->ptTypeComboBox->itemData(i, Qt::UserRole).toInt() == instFM->getPitchType(param)) {
+		if (instFM->getPitchType(param) == convertSequenceTypeForData(
+					static_cast<VisualizedInstrumentMacroEditor::SequenceType>(ui->ptTypeComboBox->itemData(i).toInt()))) {
 			ui->ptTypeComboBox->setCurrentIndex(i);
 			break;
 		}
@@ -1828,14 +1771,12 @@ void InstrumentEditorFMForm::setInstrumentPitchParameters()
 void InstrumentEditorFMForm::onPitchNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getPitchFMUsers(ui->ptNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->ptUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->ptUsersLineEdit->setText(l.join(","));
 }
 
 void InstrumentEditorFMForm::onPitchParameterChanged(int tnNum)
@@ -1857,12 +1798,14 @@ void InstrumentEditorFMForm::onPitchTypeChanged(int index)
 {
 	Q_UNUSED(index)
 
+	auto type = static_cast<VisualizedInstrumentMacroEditor::SequenceType>(ui->ptTypeComboBox->currentData(Qt::UserRole).toInt());
 	if (!isIgnoreEvent_) {
-		bt_.lock()->setPitchFMType(ui->ptNumSpinBox->value(),
-								   ui->ptTypeComboBox->currentData(Qt::UserRole).toInt());
+		bt_.lock()->setPitchFMType(ui->ptNumSpinBox->value(), convertSequenceTypeForData(type));
 		emit pitchParameterChanged(ui->ptNumSpinBox->value(), instNum_);
 		emit modified();
 	}
+
+	ui->ptEditor->setSequenceType(type);
 }
 
 void InstrumentEditorFMForm::on_ptEditGroupBox_toggled(bool arg1)
