@@ -6,6 +6,7 @@
 #include "gui/event_guard.hpp"
 #include "pitch_converter.hpp"
 #include "command_sequence.hpp"
+#include "gui/jam_layout.hpp"
 #include "misc.hpp"
 
 InstrumentEditorSSGForm::InstrumentEditorSSGForm(int num, QWidget *parent) :
@@ -18,13 +19,13 @@ InstrumentEditorSSGForm::InstrumentEditorSSGForm(int num, QWidget *parent) :
 	//========== Wave form ==========//
 	ui->waveEditor->setMaximumDisplayedRowCount(7);
 	ui->waveEditor->setDefaultRow(0);
-	ui->waveEditor->AddRow("Sq", false);
-	ui->waveEditor->AddRow("Tri", false);
-	ui->waveEditor->AddRow("Saw", false);
-	ui->waveEditor->AddRow("InvSaw", false);
-	ui->waveEditor->AddRow("SMTri", false);
-	ui->waveEditor->AddRow("SMSaw", false);
-	ui->waveEditor->AddRow("SMInvSaw", false);
+	ui->waveEditor->AddRow(tr("Sq"), false);
+	ui->waveEditor->AddRow(tr("Tri"), false);
+	ui->waveEditor->AddRow(tr("Saw"), false);
+	ui->waveEditor->AddRow(tr("InvSaw"), false);
+	ui->waveEditor->AddRow(tr("SMTri"), false);
+	ui->waveEditor->AddRow(tr("SMSaw"), false);
+	ui->waveEditor->AddRow(tr("SMInvSaw"), false);
 	ui->waveEditor->autoFitLabelWidth();
 
 	QObject::connect(ui->waveEditor, &VisualizedInstrumentMacroEditor::sequenceCommandAdded,
@@ -75,17 +76,6 @@ InstrumentEditorSSGForm::InstrumentEditorSSGForm(int num, QWidget *parent) :
 	});
 
 	//========== Tone/Noise ==========//
-	ui->tnEditor->setMaximumDisplayedRowCount(16);
-	ui->tnEditor->setDefaultRow(0);
-	ui->tnEditor->AddRow("T", false);
-	for (int i = 0; i < 32; ++i) {
-		ui->tnEditor->AddRow("N " + QString::number(i), false);
-	}
-	for (int i = 0; i < 32; ++i) {
-		ui->tnEditor->AddRow("T&N " + QString::number(i), false);
-	}
-	ui->tnEditor->autoFitLabelWidth();
-
 	QObject::connect(ui->tnEditor, &VisualizedInstrumentMacroEditor::sequenceCommandAdded,
 					 this, [&](int row, int col) {
 		if (!isIgnoreEvent_) {
@@ -138,10 +128,14 @@ InstrumentEditorSSGForm::InstrumentEditorSSGForm(int num, QWidget *parent) :
 		ui->envEditor->AddRow(QString::number(i), false);
 	}
 	for (int i = 0; i < 8; ++i) {
-		ui->envEditor->AddRow("Hard " + QString::number(i), false);
+		ui->envEditor->AddRow(tr("HEnv %1").arg(i), false);
 	}
 	ui->envEditor->autoFitLabelWidth();
 	ui->envEditor->setMultipleReleaseState(true);
+	ui->envEditor->setPermittedReleaseTypes(
+				VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE
+				| VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE
+				| VisualizedInstrumentMacroEditor::ReleaseType::FIXED);
 
 	QObject::connect(ui->envEditor, &VisualizedInstrumentMacroEditor::sequenceCommandAdded,
 					 this, [&](int row, int col) {
@@ -191,22 +185,9 @@ InstrumentEditorSSGForm::InstrumentEditorSSGForm(int num, QWidget *parent) :
 	});
 
 	//========== Arpeggio ==========//
-	ui->arpEditor->setMaximumDisplayedRowCount(15);
-	ui->arpEditor->setDefaultRow(48);
-	ui->arpEditor->setLabelDiaplayMode(true);
-	for (int i = 0; i < 96; ++i) {
-		int d = i - 48;
-		auto text = QString::number(d);
-		if (d > 0) text = "+" + text;
-		ui->arpEditor->AddRow(text, false);
-	}
-	ui->arpEditor->autoFitLabelWidth();
-	ui->arpEditor->setUpperRow(55);
-	ui->arpEditor->setMMLDisplay0As(-48);
-
-	ui->arpTypeComboBox->addItem(tr("Absolute"), 0);
-	ui->arpTypeComboBox->addItem(tr("Fix"), 1);
-	ui->arpTypeComboBox->addItem(tr("Relative"), 2);
+	ui->arpTypeComboBox->addItem(tr("Absolute"), VisualizedInstrumentMacroEditor::SequenceType::Absolute);
+	ui->arpTypeComboBox->addItem(tr("Fixed"), VisualizedInstrumentMacroEditor::SequenceType::Fixed);
+	ui->arpTypeComboBox->addItem(tr("Relative"), VisualizedInstrumentMacroEditor::SequenceType::Relative);
 
 	QObject::connect(ui->arpEditor, &VisualizedInstrumentMacroEditor::sequenceCommandAdded,
 					 this, [&](int row, int col) {
@@ -261,17 +242,14 @@ InstrumentEditorSSGForm::InstrumentEditorSSGForm(int num, QWidget *parent) :
 	ui->ptEditor->setDefaultRow(127);
 	ui->ptEditor->setLabelDiaplayMode(true);
 	for (int i = 0; i < 255; ++i) {
-		int d = i - 127;
-		auto text = QString::number(d);
-		if (d > 0) text = "+" + text;
-		ui->ptEditor->AddRow(text, false);
+		ui->ptEditor->AddRow(QString::asprintf("%+d", i - 127), false);
 	}
 	ui->ptEditor->autoFitLabelWidth();
 	ui->ptEditor->setUpperRow(134);
 	ui->ptEditor->setMMLDisplay0As(-127);
 
-	ui->ptTypeComboBox->addItem(tr("Absolute"), 0);
-	ui->ptTypeComboBox->addItem(tr("Relative"), 2);
+	ui->ptTypeComboBox->addItem(tr("Absolute"), VisualizedInstrumentMacroEditor::SequenceType::Absolute);
+	ui->ptTypeComboBox->addItem(tr("Relative"), VisualizedInstrumentMacroEditor::SequenceType::Relative);
 
 	QObject::connect(ui->ptEditor, &VisualizedInstrumentMacroEditor::sequenceCommandAdded,
 					 this, [&](int row, int col) {
@@ -352,17 +330,49 @@ void InstrumentEditorSSGForm::setColorPalette(std::shared_ptr<ColorPalette> pale
 	ui->ptEditor->setColorPalette(palette);
 }
 
+SequenceType InstrumentEditorSSGForm::convertSequenceTypeForData(VisualizedInstrumentMacroEditor::SequenceType type)
+{
+	switch (type) {
+	case VisualizedInstrumentMacroEditor::SequenceType::NoType:
+		return SequenceType::NO_SEQUENCE_TYPE;
+	case VisualizedInstrumentMacroEditor::SequenceType::Fixed:
+		return SequenceType::FIXED_SEQUENCE;
+	case VisualizedInstrumentMacroEditor::SequenceType::Absolute:
+		return SequenceType::ABSOLUTE_SEQUENCE;
+	case VisualizedInstrumentMacroEditor::SequenceType::Relative:
+		return SequenceType::RELATIVE_SEQUENCE;
+	default:
+		throw std::invalid_argument("Unexpected SequenceType.");
+	}
+}
+
+VisualizedInstrumentMacroEditor::SequenceType InstrumentEditorSSGForm::convertSequenceTypeForUI(SequenceType type)
+{
+	switch (type) {
+	case SequenceType::NO_SEQUENCE_TYPE:
+		return VisualizedInstrumentMacroEditor::SequenceType::NoType;
+	case SequenceType::FIXED_SEQUENCE:
+		return VisualizedInstrumentMacroEditor::SequenceType::Fixed;
+	case SequenceType::ABSOLUTE_SEQUENCE:
+		return VisualizedInstrumentMacroEditor::SequenceType::Absolute;
+	case SequenceType::RELATIVE_SEQUENCE:
+		return VisualizedInstrumentMacroEditor::SequenceType::Relative;
+	default:
+		throw std::invalid_argument("Unexpected SequenceType.");
+	}
+}
+
 ReleaseType InstrumentEditorSSGForm::convertReleaseTypeForData(VisualizedInstrumentMacroEditor::ReleaseType type)
 {
 	switch (type) {
 	case VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE:
-		return ReleaseType::NO_RELEASE;
-	case VisualizedInstrumentMacroEditor::ReleaseType::FIX:
-		return ReleaseType::FIX;
+		return ReleaseType::NoRelease;
+	case VisualizedInstrumentMacroEditor::ReleaseType::FIXED:
+		return ReleaseType::FixedRelease;
 	case VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE:
-		return ReleaseType::ABSOLUTE;
+		return ReleaseType::AbsoluteRelease;
 	case VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE:
-		return ReleaseType::RELATIVE;
+		return ReleaseType::RelativeRelease;
 	default:
 		throw std::invalid_argument("Unexpected ReleaseType.");
 	}
@@ -371,13 +381,13 @@ ReleaseType InstrumentEditorSSGForm::convertReleaseTypeForData(VisualizedInstrum
 VisualizedInstrumentMacroEditor::ReleaseType InstrumentEditorSSGForm::convertReleaseTypeForUI(ReleaseType type)
 {
 	switch (type) {
-	case ReleaseType::NO_RELEASE:
+	case ReleaseType::NoRelease:
 		return VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE;
-	case ReleaseType::FIX:
-		return VisualizedInstrumentMacroEditor::ReleaseType::FIX;
-	case ReleaseType::ABSOLUTE:
+	case ReleaseType::FixedRelease:
+		return VisualizedInstrumentMacroEditor::ReleaseType::FIXED;
+	case ReleaseType::AbsoluteRelease:
 		return VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE;
-	case ReleaseType::RELATIVE:
+	case ReleaseType::RelativeRelease:
 		return VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE;
 	default:
 		throw std::invalid_argument("Unexpected ReleaseType.");
@@ -432,41 +442,11 @@ void InstrumentEditorSSGForm::keyPressEvent(QKeyEvent *event)
 	default:
 		if (!event->isAutoRepeat()) {
 			// Musical keyboard
-			switch (event->key()) {
-			case Qt::Key_Z:
-			case Qt::Key_S:
-			case Qt::Key_X:
-			case Qt::Key_D:
-			case Qt::Key_C:
-			case Qt::Key_V:
-			case Qt::Key_G:
-			case Qt::Key_B:
-			case Qt::Key_H:
-			case Qt::Key_N:
-			case Qt::Key_J:
-			case Qt::Key_M:
-			case Qt::Key_Comma:
-			case Qt::Key_L:
-			case Qt::Key_Period:
-			case Qt::Key_Q:
-			case Qt::Key_2:
-			case Qt::Key_W:
-			case Qt::Key_3:
-			case Qt::Key_E:
-			case Qt::Key_R:
-			case Qt::Key_5:
-			case Qt::Key_T:
-			case Qt::Key_6:
-			case Qt::Key_Y:
-			case Qt::Key_7:
-			case Qt::Key_U:
-			case Qt::Key_I:
-			case Qt::Key_9:
-			case Qt::Key_O:
-				emit jamKeyOnEvent(event);
-				break;
-			default: break;
-			}
+			Qt::Key qtKey = static_cast<Qt::Key>(event->key());
+			try {
+				JamKey jk = getJamKeyFromLayoutMapping(qtKey, config_);
+				emit jamKeyOnEvent(jk);
+			} catch (std::invalid_argument&) {}
 		}
 		break;
 	}
@@ -477,41 +457,11 @@ void InstrumentEditorSSGForm::keyReleaseEvent(QKeyEvent *event)
 {
 	// For jam key off
 	if (!event->isAutoRepeat()) {
-		switch (event->key()) {
-		case Qt::Key_Z:
-		case Qt::Key_S:
-		case Qt::Key_X:
-		case Qt::Key_D:
-		case Qt::Key_C:
-		case Qt::Key_V:
-		case Qt::Key_G:
-		case Qt::Key_B:
-		case Qt::Key_H:
-		case Qt::Key_N:
-		case Qt::Key_J:
-		case Qt::Key_M:
-		case Qt::Key_Comma:
-		case Qt::Key_L:
-		case Qt::Key_Period:
-		case Qt::Key_Q:
-		case Qt::Key_2:
-		case Qt::Key_W:
-		case Qt::Key_3:
-		case Qt::Key_E:
-		case Qt::Key_R:
-		case Qt::Key_5:
-		case Qt::Key_T:
-		case Qt::Key_6:
-		case Qt::Key_Y:
-		case Qt::Key_7:
-		case Qt::Key_U:
-		case Qt::Key_I:
-		case Qt::Key_9:
-		case Qt::Key_O:
-			emit jamKeyOffEvent(event);
-			break;
-		default: break;
-		}
+		Qt::Key qtKey = static_cast<Qt::Key>(event->key());
+		try {
+			JamKey jk = getJamKeyFromLayoutMapping(qtKey, config_);
+			emit jamKeyOffEvent(jk);
+		} catch (std::invalid_argument&) {}
 	}
 }
 
@@ -528,9 +478,9 @@ void InstrumentEditorSSGForm::setInstrumentWaveFormParameters()
 	for (auto& com : instSSG->getWaveFormSequence()) {
 		QString str("");
 		if (isModulatedWaveFormSSG(com.type)) {
-			if (CommandSequenceUnit::isRatioData(com.data)) {
+			if (CommandSequenceUnit::checkDataType(com.data) == CommandSequenceUnit::RATIO) {
 				auto ratio = CommandSequenceUnit::data2ratio(com.data);
-				str = QString::number(ratio.first) + "/" + QString::number(ratio.second);
+				str = QString("%1/%2").arg(ratio.first).arg(ratio.second);
 			}
 			else {
 				str = QString::number(com.data);
@@ -573,14 +523,12 @@ void InstrumentEditorSSGForm::setWaveFormSequenceColumn(int col)
 void InstrumentEditorSSGForm::onWaveFormNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getWaveFormSSGUsers(ui->waveNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->waveUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->waveUsersLineEdit->setText(l.join(","));
 }
 
 void InstrumentEditorSSGForm::onWaveFormParameterChanged(int wfNum)
@@ -654,14 +602,12 @@ void InstrumentEditorSSGForm::setInstrumentToneNoiseParameters()
 void InstrumentEditorSSGForm::onToneNoiseNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getToneNoiseSSGUsers(ui->tnNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->tnUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->tnUsersLineEdit->setText(l.join((",")));
 }
 
 void InstrumentEditorSSGForm::onToneNoiseParameterChanged(int tnNum)
@@ -709,9 +655,9 @@ void InstrumentEditorSSGForm::setInstrumentEnvelopeParameters()
 	for (auto& com : instSSG->getEnvelopeSequence()) {
 		QString str("");
 		if (com.type >= 16) {
-			if (CommandSequenceUnit::isRatioData(com.data)) {
+			if (CommandSequenceUnit::checkDataType(com.data) == CommandSequenceUnit::RATIO) {
 				auto ratio = CommandSequenceUnit::data2ratio(com.data);
-				str = QString::number(ratio.first) + "/" + QString::number(ratio.second);
+				str = QString("%1/%2").arg(ratio.first).arg(ratio.second);
 			}
 			else {
 				str = QString::number(com.data);
@@ -753,14 +699,12 @@ void InstrumentEditorSSGForm::setEnvelopeSequenceColumn(int col)
 void InstrumentEditorSSGForm::onEnvelopeNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getEnvelopeSSGUsers(ui->envNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->envUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->envUsersLineEdit->setText(l.join((",")));
 }
 
 void InstrumentEditorSSGForm::onEnvelopeParameterChanged(int envNum)
@@ -820,7 +764,13 @@ void InstrumentEditorSSGForm::setInstrumentArpeggioParameters()
 	}
 	ui->arpEditor->setRelease(convertReleaseTypeForUI(instSSG->getArpeggioRelease().type),
 							  instSSG->getArpeggioRelease().begin);
-	ui->arpTypeComboBox->setCurrentIndex(instSSG->getArpeggioType());
+	for (int i = 0; i < ui->arpTypeComboBox->count(); ++i) {
+		if (instSSG->getArpeggioType() == convertSequenceTypeForData(
+					static_cast<VisualizedInstrumentMacroEditor::SequenceType>(ui->arpTypeComboBox->itemData(i).toInt()))) {
+			ui->arpTypeComboBox->setCurrentIndex(i);
+			break;
+		}
+	}
 	if (instSSG->getArpeggioEnabled()) {
 		ui->arpEditGroupBox->setChecked(true);
 		onArpeggioNumberChanged();
@@ -834,14 +784,12 @@ void InstrumentEditorSSGForm::setInstrumentArpeggioParameters()
 void InstrumentEditorSSGForm::onArpeggioNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getArpeggioSSGUsers(ui->arpNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->arpUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->arpUsersLineEdit->setText(l.join(","));
 }
 
 void InstrumentEditorSSGForm::onArpeggioParameterChanged(int tnNum)
@@ -856,30 +804,15 @@ void InstrumentEditorSSGForm::onArpeggioTypeChanged(int index)
 {
 	Q_UNUSED(index)
 
+	auto type = static_cast<VisualizedInstrumentMacroEditor::SequenceType>(
+					ui->arpTypeComboBox->currentData(Qt::UserRole).toInt());
 	if (!isIgnoreEvent_) {
-		bt_.lock()->setArpeggioSSGType(ui->arpNumSpinBox->value(),
-									   ui->arpTypeComboBox->currentData(Qt::UserRole).toInt());
+		bt_.lock()->setArpeggioSSGType(ui->arpNumSpinBox->value(), convertSequenceTypeForData(type));
 		emit arpeggioParameterChanged(ui->arpNumSpinBox->value(), instNum_);
 		emit modified();
 	}
 
-	// Update labels
-	if (index == 1) {
-		QString tn[] = { "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B" };
-		for (int i = 0; i < 8; ++i) {
-			for (int j = 0; j < 12; ++j) {
-				ui->arpEditor->setLabel(i * 12 + j, tn[j] + QString::number(i));
-			}
-		}
-	}
-	else {
-		for (int i = 0; i < 96; ++i) {
-			int d = i - 48;
-			auto text = QString::number(d);
-			if (d > 0) text = "+" + text;
-			ui->arpEditor->setLabel(i, text);
-		}
-	}
+	ui->arpEditor->setSequenceType(type);
 }
 
 void InstrumentEditorSSGForm::on_arpEditGroupBox_toggled(bool arg1)
@@ -925,7 +858,8 @@ void InstrumentEditorSSGForm::setInstrumentPitchParameters()
 	ui->ptEditor->setRelease(convertReleaseTypeForUI(instSSG->getPitchRelease().type),
 							 instSSG->getPitchRelease().begin);
 	for (int i = 0; i < ui->ptTypeComboBox->count(); ++i) {
-		if (ui->ptTypeComboBox->itemData(i, Qt::UserRole).toInt() == instSSG->getPitchType()) {
+		if (instSSG->getPitchType() == convertSequenceTypeForData(
+					static_cast<VisualizedInstrumentMacroEditor::SequenceType>(ui->ptTypeComboBox->itemData(i).toInt()))) {
 			ui->ptTypeComboBox->setCurrentIndex(i);
 			break;
 		}
@@ -943,14 +877,12 @@ void InstrumentEditorSSGForm::setInstrumentPitchParameters()
 void InstrumentEditorSSGForm::onPitchNumberChanged()
 {
 	// Change users view
-	QString str;
 	std::vector<int> users = bt_.lock()->getPitchSSGUsers(ui->ptNumSpinBox->value());
-	for (auto& n : users) {
-		str += (QString("%1").arg(n, 2, 16, QChar('0')).toUpper() + ",");
-	}
-	str.chop(1);
-
-	ui->ptUsersLineEdit->setText(str);
+	QStringList l;
+	std::transform(users.begin(), users.end(), std::back_inserter(l), [](int n) {
+		return QString("%1").arg(n, 2, 16, QChar('0')).toUpper();
+	});
+	ui->ptUsersLineEdit->setText(l.join(","));
 }
 
 void InstrumentEditorSSGForm::onPitchParameterChanged(int tnNum)
@@ -965,12 +897,14 @@ void InstrumentEditorSSGForm::onPitchTypeChanged(int index)
 {
 	Q_UNUSED(index)
 
+	auto type = static_cast<VisualizedInstrumentMacroEditor::SequenceType>(ui->ptTypeComboBox->currentData(Qt::UserRole).toInt());
 	if (!isIgnoreEvent_) {
-		bt_.lock()->setPitchSSGType(ui->ptNumSpinBox->value(),
-									ui->ptTypeComboBox->currentData(Qt::UserRole).toInt());
+		bt_.lock()->setPitchSSGType(ui->ptNumSpinBox->value(), convertSequenceTypeForData(type));
 		emit pitchParameterChanged(ui->ptNumSpinBox->value(), instNum_);
 		emit modified();
 	}
+
+	ui->ptEditor->setSequenceType(type);
 }
 
 void InstrumentEditorSSGForm::on_ptEditGroupBox_toggled(bool arg1)
