@@ -13,7 +13,6 @@
 
 VisualizedInstrumentMacroEditor::VisualizedInstrumentMacroEditor(QWidget *parent)
 	: QWidget(parent),
-	  ui(new Ui::VisualizedInstrumentMacroEditor),
 	  font_(QApplication::font()),
 	  met_(font_),
 	  maxDispRowCnt_(0),
@@ -21,14 +20,17 @@ VisualizedInstrumentMacroEditor::VisualizedInstrumentMacroEditor(QWidget *parent
 	  defaultRow_(0),
 	  hovRow_(-1),
 	  hovCol_(-1),
+	  type_(NoType),
+	  permittedReleaseType_(ReleaseType::FIXED),
+	  isLabelOmitted_(false),
+	  release_{ ReleaseType::NO_RELEASE, -1 },
+	  ui(new Ui::VisualizedInstrumentMacroEditor),
 	  pressRow_(-1),
 	  pressCol_(-1),
 	  grabLoop_(-1),
 	  isGrabLoopHead_(false),
 	  isGrabRelease_(false),
-	  release_{ VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE, -1 },
 	  isMultiReleaseState_(false),
-	  isLabelOmitted_(false),
 	  mmlBase_(0),
 	  isIgnoreEvent_(false)
 {
@@ -37,7 +39,11 @@ VisualizedInstrumentMacroEditor::VisualizedInstrumentMacroEditor(QWidget *parent
 	/* Font */
 	font_.setPointSize(10);
 	// Check font size
+#if QT_VERSION >= QT_VERSION_CHECK(5, 11, 0)
+	fontWidth_ = met_.horizontalAdvance('0');
+#else
 	fontWidth_ = met_.width('0');
+#endif
 	fontAscend_ = met_.ascent();
 	fontHeight_ = met_.height();
 	fontLeading_ = met_.leading();
@@ -114,7 +120,7 @@ void VisualizedInstrumentMacroEditor::setSequenceCommand(int row, int col, QStri
 
 	ui->panel->update();
 
-	makeMML();
+	printMML();
 
 	emit sequenceCommandChanged(row, col);
 }
@@ -127,7 +133,7 @@ void VisualizedInstrumentMacroEditor::setText(int col, QString text)
 void VisualizedInstrumentMacroEditor::setData(int col, int data)
 {
 	cols_.at(static_cast<size_t>(col)).data = data;
-	makeMML();
+	printMML();
 }
 
 int VisualizedInstrumentMacroEditor::getSequenceAt(int col) const
@@ -154,7 +160,7 @@ void VisualizedInstrumentMacroEditor::addSequenceCommand(int row, QString str, i
 
 	ui->colSizeLabel->setText(tr("Size: %1").arg(cols_.size()));
 
-	makeMML();
+	printMML();
 
 	emit sequenceCommandAdded(row, static_cast<int>(cols_.size()) - 1);
 }
@@ -186,7 +192,7 @@ void VisualizedInstrumentMacroEditor::removeSequenceCommand()
 
 	ui->colSizeLabel->setText(tr("Size: %1").arg(cols_.size()));
 
-	makeMML();
+	printMML();
 
 	emit sequenceCommandRemoved();
 }
@@ -204,26 +210,39 @@ void VisualizedInstrumentMacroEditor::addLoop(int begin, int end, int times)
 
 	loops_.insert(loops_.begin() + inx, { begin, end, times });
 
-	makeMML();
+	printMML();
 
 	onLoopChanged();
 }
 
+void VisualizedInstrumentMacroEditor::setSequenceType(SequenceType type)
+{
+	type_ = type;
+
+	updateLabels();
+	printMML();
+}
+
+void VisualizedInstrumentMacroEditor::setPermittedReleaseTypes(int types)
+{
+	permittedReleaseType_ = types;
+}
+
 void VisualizedInstrumentMacroEditor::setRelease(ReleaseType type, int point)
 {
 	release_ = { type, point };
 
-	makeMML();
+	printMML();
 }
 
 void VisualizedInstrumentMacroEditor::clearData()
 {
 	cols_.clear();
 	loops_.clear();
-	release_ = { VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE, -1 };
+	release_ = { ReleaseType::NO_RELEASE, -1 };
 	updateColumnWidth();
 
-	makeMML();
+	printMML();
 }
 
 void VisualizedInstrumentMacroEditor::clearRow()
@@ -262,8 +281,13 @@ void VisualizedInstrumentMacroEditor::setLabelDiaplayMode(bool isOmitted)
 
 void VisualizedInstrumentMacroEditor::autoFitLabelWidth()
 {
+#if QT_VERSION >= QT_VERSION_CHECK(5, 11, 0)
+	int w = met_.horizontalAdvance(VisualizedInstrumentMacroEditor::tr("Release") + " ");
+	for (auto& lab : labels_) w = std::max(w, met_.horizontalAdvance(lab + " "));
+#else
 	int w = met_.width(VisualizedInstrumentMacroEditor::tr("Release") + " ");
 	for (auto& lab : labels_) w = std::max(w, met_.width(lab + " "));
+#endif
 	labWidth_ = w;
 }
 
@@ -278,41 +302,51 @@ void VisualizedInstrumentMacroEditor::drawField()
 	QPainter painter(pixmap_.get());
 	painter.setFont(font_);
 
+	int textOffset = fontAscend_ - fontHeight_ + fontLeading_ / 2;
+
 	// Row label
 	painter.setPen(palette_->instSeqTagColor);
+	int dispCnt = getDisplayedRowCount();
 	if (isLabelOmitted_ && !labels_.empty()) {
 		painter.drawText(1,
-						 rowHeights_.front() - fontHeight_ + fontAscend_ + fontLeading_ / 2,
+						 rowHeights_.front() + textOffset,
 						 labels_[static_cast<size_t>(upperRow_)]);
-		int c = getDisplayedRowCount() / 2;
+		int c = dispCnt / 2;
 		painter.drawText(1,
-						 std::accumulate(rowHeights_.begin(), rowHeights_.begin() + c + 1, 0)
-						 - fontHeight_ + fontAscend_ + fontLeading_ / 2,
+						 std::accumulate(rowHeights_.begin(), rowHeights_.begin() + c + 1, 0) + textOffset,
 						 labels_[static_cast<size_t>(upperRow_ - c)]);
-		int l = getDisplayedRowCount() - 1;
+		int l = dispCnt - 1;
 		painter.drawText(1,
-						 std::accumulate(rowHeights_.begin(), rowHeights_.begin() + l + 1, 0)
-						 - fontHeight_ + fontAscend_ + fontLeading_ / 2,
+						 std::accumulate(rowHeights_.begin(), rowHeights_.begin() + l + 1, 0) + textOffset,
 						 labels_[static_cast<size_t>(upperRow_ - l)]);
 	}
 	else {
-		for (int i = 0; i < maxDispRowCnt_; ++i) {
+		for (int i = 0; i < dispCnt; ++i) {
 			painter.drawText(1,
-							 std::accumulate(rowHeights_.begin(), rowHeights_.begin() + i + 1, 0)
-							 - fontHeight_ + fontAscend_ + fontLeading_ / 2,
+							 std::accumulate(rowHeights_.begin(), rowHeights_.begin() + i + 1, 0) + textOffset,
 							 labels_[static_cast<size_t>(upperRow_ - i)]);
 		}
 	}
 
+	for (size_t i = 1; i < cols_.size(); i += 2) {
+		painter.fillRect(labWidth_
+						 + std::accumulate(colWidths_.begin(), colWidths_.begin() + static_cast<int>(i), 0),
+						 0,
+						 colWidths_[i], fieldHeight_,
+						 palette_->instSeqOddColColor);
+	}
+
 	// Sequence
 	painter.setPen(palette_->instSeqCellTextColor);
 	for (size_t i = 0; i < cols_.size(); ++i) {
-		if (upperRow_ >= cols_[i].row && cols_[i].row > upperRow_ - maxDispRowCnt_) {
+		if (upperRow_ >= cols_[i].row && cols_[i].row > upperRow_ - dispCnt) {
 			int x = labWidth_ + std::accumulate(colWidths_.begin(), colWidths_.begin() + static_cast<int>(i), 0);
-			int y = std::accumulate(rowHeights_.begin(), rowHeights_.begin() + (upperRow_ - cols_[i].row), 0);
-			painter.fillRect(x, y, colWidths_[i], rowHeights_[static_cast<size_t>(upperRow_ - cols_[i].row)], palette_->instSeqCellColor);
+			int hi = upperRow_ - cols_[i].row;
+			int y = std::accumulate(rowHeights_.begin(), rowHeights_.begin() + hi, 0);
+			painter.fillRect(x, y, colWidths_[i],
+							 rowHeights_[static_cast<size_t>(hi)], palette_->instSeqCellColor);
 			painter.drawText(x + 2,
-							 y + rowHeights_[static_cast<size_t>(upperRow_ - cols_[i].row)] - fontHeight_ + fontAscend_ + (fontLeading_ / 2),
+							 y + rowHeights_[static_cast<size_t>(upperRow_ - cols_[i].row)] + textOffset,
 					cols_[i].text);
 		}
 	}
@@ -330,7 +364,7 @@ void VisualizedInstrumentMacroEditor::drawLoop()
 	QPainter painter(pixmap_.get());
 	painter.setFont(font_);
 
-	painter.fillRect(0, loopY_, ui->panel->geometry().width(), fontHeight_, palette_->instSeqLoopBackColor);
+	painter.fillRect(0, loopY_, panelWidth(), fontHeight_, palette_->instSeqLoopBackColor);
 	painter.setPen(palette_->instSeqLoopTextColor);
 	painter.drawText(1, loopBaseY_, tr("Loop"));
 
@@ -352,8 +386,11 @@ void VisualizedInstrumentMacroEditor::drawLoop()
 				}
 			}
 		}
+
 		if (hovRow_ == -2 && hovCol_ == i)
-			painter.fillRect(w, loopY_, colWidths_[static_cast<size_t>(i)], fontHeight_, palette_->instSeqHovColor);
+			painter.fillRect(w, loopY_, colWidths_[static_cast<size_t>(i)],
+					fontHeight_, palette_->instSeqHovColor);
+
 		w += colWidths_[static_cast<size_t>(i)];
 	}
 }
@@ -363,7 +400,7 @@ void VisualizedInstrumentMacroEditor::drawRelease()
 	QPainter painter(pixmap_.get());
 	painter.setFont(font_);
 
-	painter.fillRect(0, releaseY_, ui->panel->geometry().width(), fontHeight_, palette_->instSeqReleaseBackColor);
+	painter.fillRect(0, releaseY_, panelWidth(), fontHeight_, palette_->instSeqReleaseBackColor);
 	painter.setPen(palette_->instSeqReleaseTextColor);
 	painter.drawText(1, releaseBaseY_, tr("Release"));
 
@@ -371,29 +408,31 @@ void VisualizedInstrumentMacroEditor::drawRelease()
 	int seqLen = static_cast<int>(cols_.size());
 	for (int i = 0; i < seqLen; ++i) {
 		if (release_.point == i) {
-			painter.fillRect(w, releaseY_, ui->panel->geometry().width() - w, fontHeight_, palette_->instSeqReleaseColor);
+			painter.fillRect(w, releaseY_, panelWidth() - w, fontHeight_, palette_->instSeqReleaseColor);
 			painter.fillRect(w, releaseY_, 2, fontHeight_, palette_->instSeqReleaseEdgeColor);
 			QString type;
 			switch (release_.type) {
-			case VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE:
+			case ReleaseType::NO_RELEASE:
 				type = "";
 				break;
-			case VisualizedInstrumentMacroEditor::ReleaseType::FIX:
-				type = tr("Fix");
+			case ReleaseType::FIXED:
+				type = tr("Fixed");
 				break;
-			case VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE:
+			case ReleaseType::ABSOLUTE:
 				type = tr("Absolute");
 				break;
-			case VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE:
+			case ReleaseType::RELATIVE:
 				type = tr("Relative");
 				break;
 			}
 			painter.setPen(palette_->instSeqReleaseTextColor);
 			painter.drawText(w + 2, releaseBaseY_, type);
 		}
+
 		if (hovRow_ == -3 && hovCol_ == i)
 			painter.fillRect(w, releaseY_, colWidths_[static_cast<size_t>(i)],
 					fontHeight_, palette_->instSeqHovColor);
+
 		w += colWidths_[static_cast<size_t>(i)];
 	}
 }
@@ -405,17 +444,21 @@ void VisualizedInstrumentMacroEditor::drawBorder()
 	painter.drawLine(labWidth_, 0, labWidth_, ui->panel->geometry().height());
 	for (int i = 1; i < maxDispRowCnt_; ++i) {
 		painter.drawLine(labWidth_, std::accumulate(rowHeights_.begin(), rowHeights_.begin() + i, 0),
-						 ui->panel->geometry().width(), std::accumulate(rowHeights_.begin(), rowHeights_.begin() + i, 0));
+						 panelWidth(), std::accumulate(rowHeights_.begin(), rowHeights_.begin() + i, 0));
 	}
 }
 
 void VisualizedInstrumentMacroEditor::drawShadow()
 {
 	QPainter painter(pixmap_.get());
-	painter.fillRect(0, 0, ui->panel->geometry().width(), ui->panel->geometry().height(), palette_->instSeqMaskColor);
+	painter.fillRect(0, 0, panelWidth(), ui->panel->geometry().height(), palette_->instSeqMaskColor);
 }
 
-void VisualizedInstrumentMacroEditor::makeMML()
+void VisualizedInstrumentMacroEditor::updateLabels()
+{
+}
+
+void VisualizedInstrumentMacroEditor::printMML()
 {
 	if (cols_.empty()) return;
 
@@ -426,7 +469,7 @@ void VisualizedInstrumentMacroEditor::makeMML()
 	for (int cnt = 0; cnt < seqLen; ++cnt) {
 		if (release_.point == cnt) {
 			switch (release_.type) {
-			case ReleaseType::FIX:		text += "| ";	break;
+			case ReleaseType::FIXED:	text += "| ";	break;
 			case ReleaseType::ABSOLUTE:	text += "/ ";	break;
 			case ReleaseType::RELATIVE:	text += ": ";	break;
 			default:									break;
@@ -443,7 +486,7 @@ void VisualizedInstrumentMacroEditor::makeMML()
 			}
 		}
 
-		text += (QString::number(cols_[static_cast<size_t>(cnt)].row + mmlBase_) + " ");
+		text += (convertSequenceDataUnitToMML(cols_[static_cast<size_t>(cnt)]) + " ");
 
 		while (!lstack.empty()) {
 			if (lstack.back().end == cnt) {
@@ -463,6 +506,121 @@ void VisualizedInstrumentMacroEditor::makeMML()
 	ui->lineEdit->setText(text);
 }
 
+QString VisualizedInstrumentMacroEditor::convertSequenceDataUnitToMML(Column col)
+{
+	return QString::number(col.row + mmlBase_);
+}
+
+void VisualizedInstrumentMacroEditor::interpretMML()
+{
+	if (cols_.empty()) return;
+
+	QString text = ui->lineEdit->text();
+
+	std::vector<Column> column;
+	std::vector<Loop> loop;
+	std::vector<size_t> lstack;
+	Release release = { ReleaseType::NO_RELEASE, -1 };
+
+	int cnt = 0;
+	while (!text.isEmpty()) {
+
+		QRegularExpressionMatch m = QRegularExpression("^\\[").match(text);
+		if (m.hasMatch()) {
+			loop.push_back({ cnt, cnt, 1 });
+			lstack.push_back(loop.size() - 1);
+			text.remove(QRegularExpression("^\\["));
+			continue;
+		}
+
+		m = QRegularExpression("^\\](\\d*)").match(text);
+		if (m.hasMatch()) {
+			if (lstack.empty() || cnt == 0) return;
+			loop[lstack.back()].end = cnt - 1;
+			if (!m.captured(1).isEmpty()) {
+				int t = m.captured(1).toInt();
+				if (t > 1) loop[lstack.back()].times = t;
+				else return;
+			}
+			lstack.pop_back();
+			text.remove(QRegularExpression("^\\]\\d*"));
+			continue;
+		}
+
+		if (permittedReleaseType_ & ReleaseType::FIXED) {
+			m = QRegularExpression("^\\|").match(text);
+			if (m.hasMatch()) {
+				if (release.point > -1) return;
+				release = { ReleaseType::FIXED, cnt };
+				text.remove(QRegularExpression("^\\|"));
+				continue;
+			}
+		}
+
+		if (permittedReleaseType_ & ReleaseType::ABSOLUTE) {
+			m = QRegularExpression("^/").match(text);
+			if (m.hasMatch()) {
+				if (release.point > -1) return;
+				release = { ReleaseType::ABSOLUTE, cnt };
+				text.remove(QRegularExpression("^/"));
+				continue;
+			}
+		}
+
+		if (permittedReleaseType_ & ReleaseType::RELATIVE) {
+			m = QRegularExpression("^:").match(text);
+			if (m.hasMatch()) {
+				if (release.point > -1) return;
+				release = { ReleaseType::RELATIVE, cnt };
+				text.remove(QRegularExpression("^:"));
+				continue;
+			}
+		}
+
+		if (interpretDataInMML(text, cnt, column)) continue;	// Data
+
+		m = QRegularExpression("^ +").match(text);
+		if (m.hasMatch()) {
+			text.remove(QRegularExpression("^ +"));
+			continue;
+		}
+
+		return;
+	}
+
+	if (column.empty()) return;
+	if (!lstack.empty()) return;
+	if (release.point > -1 && release.point >= static_cast<int>(column.size())) return;
+
+	while (cols_.size() > 1) removeSequenceCommand();
+	setSequenceCommand(column.front().row, 0);
+	for (size_t i = 1; i < column.size(); ++i) {
+		addSequenceCommand(column[i].row);
+	}
+
+	loops_ = loop;
+	onLoopChanged();
+
+	release_ = release;
+	emit releaseChanged(release.type, release.point);
+
+	ui->panel->update();
+}
+
+bool VisualizedInstrumentMacroEditor::interpretDataInMML(QString& text, int& cnt, std::vector<Column>& column)
+{
+	QRegularExpressionMatch m = QRegularExpression("^(-?\\d+)").match(text);
+	if (m.hasMatch()) {
+		int d = m.captured(1).toInt() - mmlBase_;
+		if (d < 0 || maxInMML() <= d) return false;
+		column.push_back({ d, -1, "" });
+		++cnt;
+		text.remove(QRegularExpression("^-?\\d+"));
+		return true;
+	}
+	return false;
+}
+
 int VisualizedInstrumentMacroEditor::checkLoopRegion(int col)
 {
 	int ret = -1;
@@ -523,7 +681,7 @@ void VisualizedInstrumentMacroEditor::moveLoop()
 		}
 	}
 
-	makeMML();
+	printMML();
 }
 
 void VisualizedInstrumentMacroEditor::setMMLDisplay0As(int n)
@@ -531,10 +689,15 @@ void VisualizedInstrumentMacroEditor::setMMLDisplay0As(int n)
 	mmlBase_ = n;
 }
 
+int VisualizedInstrumentMacroEditor::panelWidth() const
+{
+	return ui->panel->geometry().width();
+}
+
 /********** Events **********/
 bool VisualizedInstrumentMacroEditor::eventFilter(QObject*object, QEvent* event)
 {
-	if (object->objectName() == "panel") {
+	if (object == ui->panel) {
 		switch (event->type()) {
 		case QEvent::Paint:
 			paintEventInView(dynamic_cast<QPaintEvent*>(event));
@@ -609,6 +772,12 @@ void VisualizedInstrumentMacroEditor::resizeEventInView(QResizeEvent* event)
 	initDisplay();
 }
 
+int VisualizedInstrumentMacroEditor::detectRowNumberForMouseEvent(int col, int internalRow) const
+{
+	Q_UNUSED(col)
+	return upperRow_ - internalRow;
+}
+
 void VisualizedInstrumentMacroEditor::mousePressEventInView(QMouseEvent* event)
 {
 	if (!cols_.size()) return;
@@ -671,7 +840,7 @@ void VisualizedInstrumentMacroEditor::mousePressEventInView(QMouseEvent* event)
 					}
 					else {	// Loop count up
 						++loops_[static_cast<size_t>(i)].times;
-						makeMML();
+						printMML();
 						onLoopChanged();
 					}
 					break;
@@ -685,7 +854,7 @@ void VisualizedInstrumentMacroEditor::mousePressEventInView(QMouseEvent* event)
 						else {	// Erase loop
 							loops_.erase(loops_.begin() + i);
 						}
-						makeMML();
+						printMML();
 						onLoopChanged();
 					}
 					break;
@@ -701,27 +870,27 @@ void VisualizedInstrumentMacroEditor::mousePressEventInView(QMouseEvent* event)
 				case Qt::LeftButton:
 				{
 					if (release_.point == -1 || pressCol_ < release_.point) {	// New release
-						release_.type = (release_.type == VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE)
-										? VisualizedInstrumentMacroEditor::ReleaseType::FIX
+						release_.type = (release_.type == ReleaseType::NO_RELEASE)
+										? ReleaseType::FIXED
 										: release_.type;
 						release_.point = pressCol_;
-						makeMML();
+						printMML();
 						emit releaseChanged(release_.type, release_.point);
 					}
 					else if (isMultiReleaseState_) {	// Change release type
 						switch (release_.type) {
-						case VisualizedInstrumentMacroEditor::ReleaseType::FIX:
-							release_.type = VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE;
+						case ReleaseType::FIXED:
+							release_.type = ReleaseType::ABSOLUTE;
 							break;
-						case VisualizedInstrumentMacroEditor::ReleaseType::ABSOLUTE:
-							release_.type = VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE;
+						case ReleaseType::ABSOLUTE:
+							release_.type = ReleaseType::RELATIVE;
 							break;
-						case VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE:
-						case VisualizedInstrumentMacroEditor::ReleaseType::RELATIVE:
-							release_.type = VisualizedInstrumentMacroEditor::ReleaseType::FIX;
+						case ReleaseType::NO_RELEASE:
+						case ReleaseType::RELATIVE:
+							release_.type = ReleaseType::FIXED;
 							break;
 						}
-						makeMML();
+						printMML();
 						emit releaseChanged(release_.type, release_.point);
 					}
 					break;
@@ -729,8 +898,8 @@ void VisualizedInstrumentMacroEditor::mousePressEventInView(QMouseEvent* event)
 				case Qt::RightButton:
 				{
 					if (pressCol_ >= release_.point) {	// Erase release
-						release_ = { VisualizedInstrumentMacroEditor::ReleaseType::NO_RELEASE, -1 };
-						makeMML();
+						release_ = { ReleaseType::NO_RELEASE, -1 };
+						printMML();
 						emit releaseChanged(release_.type, release_.point);
 					}
 					break;
@@ -741,7 +910,9 @@ void VisualizedInstrumentMacroEditor::mousePressEventInView(QMouseEvent* event)
 			}
 		}
 		else {
-			setSequenceCommand(upperRow_ - hovRow_, hovCol_);
+			setSequenceCommand(detectRowNumberForMouseEvent(hovCol_, hovRow_), hovCol_);
+			prevPressRow_ = hovRow_;
+			prevPressCol_ = hovCol_;
 		}
 	}
 
@@ -762,7 +933,7 @@ void VisualizedInstrumentMacroEditor::mouseReleaseEventInView(QMouseEvent* event
 		if (event->button() == Qt::LeftButton) {
 			if (hovCol_ > -1) {
 				release_.point = hovCol_;
-				makeMML();
+				printMML();
 				emit releaseChanged(release_.type, release_.point);
 			}
 		}
@@ -770,6 +941,8 @@ void VisualizedInstrumentMacroEditor::mouseReleaseEventInView(QMouseEvent* event
 
 	pressRow_ = -1;
 	pressCol_ = -1;
+	prevPressRow_ = -1;
+	prevPressCol_ = -1;
 	grabLoop_ = -1;
 	isGrabLoopHead_ = false;
 	isGrabRelease_ = false;
@@ -782,8 +955,11 @@ void VisualizedInstrumentMacroEditor::mouseMoveEventInView()
 	if (!cols_.size()) return;
 
 	if (pressRow_ >= 0 && pressCol_ >= 0 && hovRow_ >= 0 && hovCol_ >= 0) {
-		if (cols_[static_cast<size_t>(upperRow_ - hovCol_)].row != (upperRow_ - hovRow_))
-			setSequenceCommand(upperRow_ - hovRow_, hovCol_);
+		if (prevPressRow_ != hovRow_ || prevPressCol_ != hovCol_) {
+			prevPressRow_ = hovRow_;
+			prevPressCol_ = hovCol_;
+			setSequenceCommand(detectRowNumberForMouseEvent(hovCol_, hovRow_), hovCol_);
+		}
 	}
 }
 
@@ -874,105 +1050,15 @@ void VisualizedInstrumentMacroEditor::on_verticalScrollBar_valueChanged(int valu
 	}
 }
 
+int VisualizedInstrumentMacroEditor::maxInMML() const
+{
+	return static_cast<int>(labels_.size());
+}
+
 void VisualizedInstrumentMacroEditor::on_lineEdit_editingFinished()
 {
-	if (cols_.empty()) return;
-
-	QString text = ui->lineEdit->text();
-
-	std::vector<Column> column;
-	std::vector<Loop> loop;
-	std::vector<size_t> lstack;
-	Release release = { ReleaseType::NO_RELEASE, -1 };
-
-	int cnt = 0;
-	int labCnt = static_cast<int>(labels_.size());
-	while (!text.isEmpty()) {
-
-		QRegularExpressionMatch m = QRegularExpression("^\\[").match(text);
-		if (m.hasMatch()) {
-			loop.push_back({ cnt, cnt, 1 });
-			lstack.push_back(loop.size() - 1);
-			text.remove(QRegularExpression("^\\["));
-			continue;
-		}
-
-		m = QRegularExpression("^\\](\\d*)").match(text);
-		if (m.hasMatch()) {
-			if (lstack.empty() || cnt == 0) return;
-			loop[lstack.back()].end = cnt - 1;
-			if (!m.captured(1).isEmpty()) {
-				int t = m.captured(1).toInt();
-				if (t > 1) loop[lstack.back()].times = t;
-				else return;
-			}
-			lstack.pop_back();
-			text.remove(QRegularExpression("^\\]\\d*"));
-			continue;
-		}
-
-		m = QRegularExpression("^\\|").match(text);
-		if (m.hasMatch()) {
-			if (release.point > -1) return;
-			release = { ReleaseType::FIX, cnt };
-			text.remove(QRegularExpression("^\\|"));
-			continue;
-		}
-
-		m = QRegularExpression("^/").match(text);
-		if (m.hasMatch()) {
-			if (release.point > -1) return;
-			release = { ReleaseType::ABSOLUTE, cnt };
-			text.remove(QRegularExpression("^/"));
-			continue;
-		}
-
-		m = QRegularExpression("^:").match(text);
-		if (m.hasMatch()) {
-			if (release.point > -1) return;
-			release = { ReleaseType::RELATIVE, cnt };
-			text.remove(QRegularExpression("^:"));
-			continue;
-		}
-
-		m = QRegularExpression("^(-?\\d+)").match(text);
-		if (m.hasMatch()) {
-			int d = m.captured(1).toInt() - mmlBase_;
-			if (d < 0 || labCnt <= d) return;
-			column.push_back({ d, -1, "" });
-			++cnt;
-			text.remove(QRegularExpression("^-?\\d+"));
-			continue;
-		}
-
-		m = QRegularExpression("^ +").match(text);
-		if (m.hasMatch()) {
-			text.remove(QRegularExpression("^ +"));
-			continue;
-		}
-
-		return;
-	}
-
-	if (column.empty()) return;
-	if (!lstack.empty()) return;
-	if (release.point > -1 && release.point >= static_cast<int>(column.size())) return;
-
-	while (cols_.size() > 1) removeSequenceCommand();
-	setSequenceCommand(column.front().row, 0);
-	for (size_t i = 1; i < column.size(); ++i) {
-		addSequenceCommand(column[i].row);
-	}
-
-	loops_ = loop;
-	onLoopChanged();
-
-	release_ = release;
-	emit releaseChanged(release.type, release.point);
-
-	ui->panel->update();
-
-	makeMML();
+	interpretMML();
+	printMML();
 }
 
 void VisualizedInstrumentMacroEditor::onLoopChanged()
@@ -993,7 +1079,7 @@ void VisualizedInstrumentMacroEditor::updateColumnWidth()
 
 	if (!cols_.size()) return;
 
-	float ww = (ui->panel->geometry().width() - labWidth_) / static_cast<float>(cols_.size());
+	float ww = (panelWidth() - labWidth_) / static_cast<float>(cols_.size());
 	int w = static_cast<int>(ww);
 	float dif = ww - w;
 	float sum = 0;
