@@ -11,6 +11,7 @@
 #include <QCloseEvent>
 #include <QDragEnterEvent>
 #include <QDropEvent>
+#include <QShowEvent>
 #include <QMessageBox>
 #include <QResizeEvent>
 #include <QMoveEvent>
@@ -27,6 +28,7 @@
 #include "gui/file_history.hpp"
 #include "gui/effect_list_dialog.hpp"
 #include "gui/keyboard_shortcut_list_dialog.hpp"
+#include "file_io_error.hpp"
 
 class AbstractBank;
 
@@ -44,6 +46,7 @@ public:
 
 protected:
 	bool eventFilter(QObject *watched, QEvent *event) override;
+	void showEvent(QShowEvent* event) override;
 	void keyPressEvent(QKeyEvent* event) override;
 	void keyReleaseEvent(QKeyEvent* event) override;
 	void dragEnterEvent(QDragEnterEvent* event) override;
@@ -78,7 +81,7 @@ private:
 	void removeInstrument(int row);
 	void editInstrument();
 	int findRowFromInstrumentList(int instNum);
-	void editInstrumentName();
+	void renameInstrument();
 	void cloneInstrument();
 	void deepCloneInstrument();
 	void loadInstrument();
@@ -104,7 +107,8 @@ private:
 	void startPlayPattern();
 	void startPlayFromCurrentStep();
 	void stopPlaySong();
-	void lockControls(bool isLock);
+	bool hasLockedWigets_;
+	void lockWidgets(bool isLock);
 
 	// Octave change
 	void changeOctave(bool upFlag);
@@ -112,12 +116,13 @@ private:
 	// Configuration change
 	void changeConfiguration();
 	void setMidiConfiguration();
+	void updateFonts();
 
 	// History change
 	void changeFileHistory(QString file);
 
-	// Layout decypherer
-	JamKey getJamKeyFromLayoutMapping(Qt::Key key);
+	// Backup
+	bool backupModule(QString srcFile);
 
 	void setWindowTitle();
 	void setModifiedTrue();
@@ -125,11 +130,17 @@ private:
 	QString getModuleFileBaseName() const;
 	int getSelectedFileFilter(QString& file, QStringList& filters) const;
 
+	void showFileIOErrorDialog(const FileIOError& e, const QString sub = "");
+
 	bool isEditedPattern_, isEditedOrder_, isEditedInstList_;
 	bool isSelectedPO_;
 
+	bool hasShownOnce_;
+
 	bool isSavedModBefore_;
 
+	bool firstViewUpdateRequest_;
+
 	// Sub tool bar
 	QSpinBox* octave_;
 	QSpinBox *highlight1_, *highlight2_;
@@ -140,6 +151,7 @@ private:
 	QLabel* statusInst_;
 	QLabel* statusOctave_;
 	QLabel* statusIntr_;
+	QLabel* statusMixer_;
 	QLabel* statusBpm_;
 	QLabel* statusPlayPos_;
 
@@ -152,6 +164,9 @@ private:
 	int midiKeyEventMethod_;
 	int midiProgramEventMethod_;
 
+	void updateInstrumentListColors();
+	void freezeViews();
+
 private slots:
 	void on_instrumentListWidget_customContextMenuRequested(const QPoint &pos);
 	void on_instrumentListWidget_itemDoubleClicked(QListWidgetItem *item);
@@ -225,21 +240,33 @@ private slots:
 	void on_actionS98_triggered();
 	void on_actionMix_triggered();
 	void on_actionOverwrite_triggered();
-	void onNewTickSignaled();
+	void onNewTickSignaledRealChip();
+	void onNewTickSignaled(int state);
 	void on_actionClear_triggered();
 	void on_keyRepeatCheckBox_stateChanged(int arg1);
 	void updateVisuals();
 	void on_action_Effect_List_triggered();
 	void on_actionShortcuts_triggered();
 	void on_actionExport_To_Bank_File_triggered();
+	void on_actionE_xpand_Effect_Column_triggered();
+	void on_actionS_hrink_Effect_Column_triggered();
+	void on_actionRemove_Duplicate_Instruments_triggered();
+	void on_actionRename_Instrument_triggered();
 
 	inline bool showUndoResetWarningDialog(QString text)
 	{
-		return (QMessageBox::warning(this, "Warning",
-									 text + " If you execute this command, the command history is reset.",
+		return (QMessageBox::warning(this, tr("Warning"),
+									 tr("%1 If you execute this command, the command history is reset.").arg(text),
 									 QMessageBox::Yes | QMessageBox::No,
 									 QMessageBox::No) == QMessageBox::Yes);
 	}
+
+	inline void showStreamFailedDialog()
+	{
+		QMessageBox::critical(this, tr("Error"),
+							  tr("Could not open the audio stream. Please change the sound settings in Configuration."),
+							  QMessageBox::Ok, QMessageBox::Ok);
+	}
 };
 
 #endif // MAINWINDOW_HPP
