@@ -1,11 +1,11 @@
 #include "command_sequence.hpp"
 
-CommandSequence::CommandSequence(int num, int seqType, int comType, int comData)
+CommandSequence::CommandSequence(int num, SequenceType seqType, int comType, int comData)
 	: AbstractInstrumentProperty(num),
 	  DEF_COM_TYPE(comType),
 	  DEF_COM_DATA(comData),
 	  type_(seqType),
-	  release_{ ReleaseType::NO_RELEASE, -1 }
+	  release_{ ReleaseType::NoRelease, -1 }
 {
 	seq_.push_back({ comType, comData });
 }
@@ -14,12 +14,18 @@ CommandSequence::CommandSequence(const CommandSequence& other)
 	: AbstractInstrumentProperty(other),
 	  DEF_COM_TYPE(other.DEF_COM_TYPE),
 	  DEF_COM_DATA(other.DEF_COM_DATA),
+	  type_(other.type_),
 	  seq_(other.seq_),
 	  loops_(other.loops_),
 	  release_(other.release_)
 {
 }
 
+bool operator==(const CommandSequence& a, const CommandSequence& b)
+{
+	return (a.type_ == b.type_ && a.seq_ == b.seq_ && a.loops_ == b.loops_ && a.release_ == b.release_);
+}
+
 std::unique_ptr<CommandSequence> CommandSequence::clone()
 {
 	std::unique_ptr<CommandSequence> clone = std::make_unique<CommandSequence>(*this);
@@ -27,12 +33,12 @@ std::unique_ptr<CommandSequence> CommandSequence::clone()
 	return clone;
 }
 
-void CommandSequence::setType(int type)
+void CommandSequence::setType(SequenceType type)
 {
 	type_ = type;
 }
 
-int CommandSequence::getType() const
+SequenceType CommandSequence::getType() const
 {
 	return type_;
 }
@@ -198,12 +204,12 @@ int CommandSequence::Iterator::next(bool isReleaseBegin)
 		loopStack_.clear();
 		isRelease_ = true;
 		switch (seq_->release_.type) {
-		case ReleaseType::NO_RELEASE:
+		case ReleaseType::NoRelease:
 			break;
-		case ReleaseType::FIX:
+		case ReleaseType::FixedRelease:
 			next = seq_->release_.begin;
 			break;
-		case ReleaseType::ABSOLUTE:
+		case ReleaseType::AbsoluteRelease:
 		{
 			int crtr;
 			if (pos_ == -1) {
@@ -228,7 +234,7 @@ int CommandSequence::Iterator::next(bool isReleaseBegin)
 			}
 			break;
 		}
-		case ReleaseType::RELATIVE:
+		case ReleaseType::RelativeRelease:
 		{
 			if (pos_ == -1) {
 				int prevIdx = seq_->release_.begin - 1;
