@@ -2,15 +2,22 @@
 
 #include <cstdint>
 #include <memory>
-#include <map>
+#include <unordered_map>
 #include <deque>
 #include "opna.hpp"
 #include "instrument.hpp"
 #include "effect_iterator.hpp"
 #include "chips/chip_misc.h"
 #include "chips/scci/scci.h"
+#include "enum_hash.hpp"
 #include "misc.hpp"
 
+struct RegisterUnit
+{
+	int address, value;
+	bool hasCompleted;
+};
+
 struct ToneDetail
 {
 	int octave;
@@ -24,6 +31,12 @@ struct ToneNoise
 	int noisePeriod_;
 };
 
+struct WaveForm
+{
+	SSGWaveFormType type;
+	int data;	// Same format with CommandSequenceUnit::data
+};
+
 class OPNAController
 {
 public:
@@ -35,6 +48,13 @@ public:
 	// Forward instrument sequence
 	void tickEvent(SoundSource src, int ch);
 
+	// Direct register set
+	void sendRegisterAddress(int bank, int address);
+	void sendRegisterValue(int value);
+
+	// Update register states after tick process
+	void updateRegisterStates();
+
 	// Stream type
 	void useSCCI(SoundInterfaceManager* manager);
 	bool isUsedSCCI() const;
@@ -43,7 +63,7 @@ public:
 	void getStreamSamples(int16_t* container, size_t nSamples);
 	void getOutputHistory(int16_t* history);
 
-	enum { OutputHistorySize = 1024 };
+	static constexpr int OUTPUT_HISTORY_SIZE = 1024;
 
 	// Chip mode
 	void setMode(SongType mode);
@@ -63,6 +83,8 @@ private:
 	std::unique_ptr<chip::OPNA> opna_;
 	SongType mode_;
 
+	std::vector<RegisterUnit> registerSetBuf_;
+
 	void initChip();
 
 	void fillOutputHistory(const int16_t* outputs, size_t nSamples);
@@ -74,7 +96,6 @@ public:
 	void keyOnFM(int ch, Note note, int octave, int pitch, bool isJam = false);
 	void keyOnFM(int ch, int echoBuf);
 	void keyOffFM(int ch, bool isJam = false);
-	void resetFMChannelEnvelope(int ch);
 	void updateEchoBufferFM(int ch, int octave, Note note, int pitch);
 
 	// Set Instrument
@@ -83,6 +104,8 @@ public:
 	void updateInstrumentFMEnvelopeParameter(int envNum, FMEnvelopeParameter param);
 	void setInstrumentFMOperatorEnabled(int envNum, int opNum);
 	void updateInstrumentFMLFOParameter(int lfoNum, FMLFOParameter param);
+	void resetFMChannelEnvelope(int ch);
+	void restoreFMEnvelopeFromReset(int ch);
 
 	// Set volume
 	void setVolumeFM(int ch, int volume);
@@ -101,6 +124,13 @@ public:
 	void setDetuneFM(int ch, int pitch);
 	void setNoteSlideFM(int ch, int speed, int seminote);
 	void setTransposeEffectFM(int ch, int seminote);
+	void setFBControlFM(int ch, int value);
+	void setTLControlFM(int ch, int op, int value);
+	void setMLControlFM(int ch, int op, int value);
+	void setARControlFM(int ch, int op, int value);
+	void setDRControlFM(int ch, int op, int value);
+	void setRRControlFM(int ch, int op, int value);
+	void setBrightnessFM(int ch, int value);
 
 	// For state retrieve
 	void haltSequencesFM(int ch);
@@ -128,12 +158,12 @@ private:
 	/// bit1: left on/off
 	uint8_t panFM_[6];
 	bool isMuteFM_[9];
-	bool enableEnvResetFM_[9];
+	bool enableEnvResetFM_[9], hasResetEnvFM_[9];
 	int lfoFreq_;
 	int lfoStartCntFM_[6];
 	bool hasPreSetTickEventFM_[9];
 	bool needToneSetFM_[9];
-	std::map<FMEnvelopeParameter, std::unique_ptr<CommandSequence::Iterator>> opSeqItFM_[6];
+	std::unordered_map<FMEnvelopeParameter, std::unique_ptr<CommandSequence::Iterator>> opSeqItFM_[6];
 	std::unique_ptr<SequenceIteratorInterface> arpItFM_[9];
 	std::unique_ptr<CommandSequence::Iterator> ptItFM_[9];
 	bool isArpEffFM_[9];
@@ -147,6 +177,9 @@ private:
 	int sumNoteSldFM_[9];
 	bool noteSldFMSetFlag_[9];
 	int transposeFM_[9];
+	bool isFBCtrlFM_[6], isTLCtrlFM_[6][4], isMLCtrlFM_[6][4], isARCtrlFM_[6][4];
+	bool isDRCtrlFM_[6][4], isRRCtrlFM_[6][4];
+	bool isBrightFM_[6][4];
 
 	void initFM();
 
@@ -178,7 +211,63 @@ private:
 
 	void setInstrumentFMProperties(int ch);
 
-	bool isCareer(int op, int al);
+	static bool isCarrier(int op, int al);
+	static std::vector<int> getOperatorsInLevel(int level, int al);
+
+	inline FMEnvelopeParameter getParameterTL(int op) const
+	{
+		switch (op) {
+		case 0:	return FMEnvelopeParameter::TL1;
+		case 1:	return FMEnvelopeParameter::TL2;
+		case 2:	return FMEnvelopeParameter::TL3;
+		case 3:	return FMEnvelopeParameter::TL4;
+		default:	throw std::invalid_argument("Invalid operator value.");
+		}
+	}
+
+	inline FMEnvelopeParameter getParameterML(int op) const
+	{
+		switch (op) {
+		case 0:	return FMEnvelopeParameter::ML1;
+		case 1:	return FMEnvelopeParameter::ML2;
+		case 2:	return FMEnvelopeParameter::ML3;
+		case 3:	return FMEnvelopeParameter::ML4;
+		default:	throw std::invalid_argument("Invalid operator value.");
+		}
+	}
+
+	inline FMEnvelopeParameter getParameterAR(int op) const
+	{
+		switch (op) {
+		case 0:	return FMEnvelopeParameter::AR1;
+		case 1:	return FMEnvelopeParameter::AR2;
+		case 2:	return FMEnvelopeParameter::AR3;
+		case 3:	return FMEnvelopeParameter::AR4;
+		default:	throw std::invalid_argument("Invalid operator value.");
+		}
+	}
+
+	inline FMEnvelopeParameter getParameterDR(int op) const
+	{
+		switch (op) {
+		case 0:	return FMEnvelopeParameter::DR1;
+		case 1:	return FMEnvelopeParameter::DR2;
+		case 2:	return FMEnvelopeParameter::DR3;
+		case 3:	return FMEnvelopeParameter::DR4;
+		default:	throw std::invalid_argument("Invalid operator value.");
+		}
+	}
+
+	inline FMEnvelopeParameter getParameterRR(int op) const
+	{
+		switch (op) {
+		case 0:	return FMEnvelopeParameter::RR1;
+		case 1:	return FMEnvelopeParameter::RR2;
+		case 2:	return FMEnvelopeParameter::RR3;
+		case 3:	return FMEnvelopeParameter::RR4;
+		default:	throw std::invalid_argument("Invalid operator value.");
+		}
+	}
 
 	inline uint8_t calculateTL(int ch, uint8_t data) const
 	{
@@ -212,6 +301,10 @@ public:
 	void setDetuneSSG(int ch, int pitch);
 	void setNoteSlideSSG(int ch, int speed, int seminote);
 	void setTransposeEffectSSG(int ch, int seminote);
+	void setToneNoiseMixSSG(int ch, int value);
+	void setNoisePitchSSG(int ch, int pitch);
+	void setHardEnvelopePeriod(int ch, bool high, int period);
+	void setAutoEnvelopeSSG(int ch, int shift, int shape);
 
 	// For state retrieve
 	void haltSequencesSSG(int ch);
@@ -239,15 +332,12 @@ private:
 	bool isMuteSSG_[3];
 	bool hasPreSetTickEventSSG_[3];
 	bool needEnvSetSSG_[3];
+	bool setHardEnvIfNecessary_[3];
 	bool needMixSetSSG_[3];
 	bool needToneSetSSG_[3];
 	bool needSqMaskFreqSetSSG_[3];
 	std::unique_ptr<CommandSequence::Iterator> wfItSSG_[3];
-	struct
-	{
-		SSGWaveFormType type;
-		int data;
-	} wfSSG_[3];
+	WaveForm wfSSG_[3];
 	std::unique_ptr<CommandSequence::Iterator> envItSSG_[3];
 	CommandSequenceUnit envSSG_[3];
 	std::unique_ptr<CommandSequence::Iterator> tnItSSG_[3];
@@ -264,6 +354,9 @@ private:
 	int sumNoteSldSSG_[3];
 	bool noteSldSSGSetFlag_;
 	int transposeSSG_[3];
+	int toneNoiseMixSSG_[3];
+	int noisePitchSSG_;
+	int hardEnvPeriodHighSSG_, hardEnvPeriodLowSSG_;
 	std::unique_ptr<int16_t[]> outputHistory_;
 	size_t outputHistoryIndex_;
 	std::unique_ptr<int16_t[]> outputHistoryReady_;
@@ -297,8 +390,8 @@ private:
 	/*----- Drum -----*/
 public:
 	// Key on-off
-	void keyOnDrum(int ch);
-	void keyOffDrum(int ch);
+	void setKeyOnFlagDrum(int ch);
+	void setKeyOffFlagDrum(int ch);
 
 	// Set volume
 	void setVolumeDrum(int ch, int volume);
@@ -313,6 +406,7 @@ public:
 	bool isMuteDrum(int ch);
 
 private:
+	uint8_t keyOnFlagDrum_, keyOffFlagDrum_;
 	int volDrum_[6], mVolDrum_, tmpVolDrum_[6];
 	/// bit0: right on/off
 	/// bit1: left on/off
@@ -320,4 +414,6 @@ private:
 	bool isMuteDrum_[6];
 
 	void initDrum();
+
+	void updateKeyOnOffStatusDrum();
 };
